"""This module/file contains various implementations of greedy searches of topology actions.

"""


import logging
from random import randint
from typing import List, Tuple

import numpy as np
from grid2op.Action import BaseAction
from grid2op.Environment import BaseEnv
from grid2op.Observation import BaseObservation
from tqdm import tqdm

from curriculumagent.common.utilities import find_best_line_to_reconnect, is_valid
from curriculumagent.teacher.submodule.common import affected_substations
from curriculumagent.teacher.submodule.encoded_action import EncodedTopologyAction


def topology_search_topk(
        env: BaseEnv, obs: BaseObservation, all_actions: List[BaseAction], top_k=1000
) -> List[Tuple[float, BaseAction]]:
    """Perform a search over all topology actions of the given environment and return the top_k best ones, sorted by
    their rho improvement.

    Args:
        env: The environment to search in.
        obs: The current observation of the given environment.
        all_actions: All actions to simulate and rank.
        top_k: How many final actions should be in the returned list.

    Returns:
        A list containing tuples of the rho_improvement and action.
    """
    min_rho, overflow_id = obs.rho.max(), obs.rho.argmax()
    old_max_rho = min_rho  # save old max_rho for comparison

    performed_actions = []

    for action in all_actions:
        valid, reason = env._game_rules(action, env)  # noqa
        if not valid:
            logging.debug(f"Found action is not valid: {reason}")
            continue

        obs_, _, done, _ = obs.simulate(action)
        if not done:
            new_max_rho = obs_.rho.max()
            rho_improvement = old_max_rho - new_max_rho
            performed_actions.append((rho_improvement, action))

    best_actions = sorted(performed_actions, key=lambda tup: tup[0], reverse=True)[:top_k]

    return best_actions


def topology_search_tuples(
        env: BaseEnv,
        best_unitary_actions: List[BaseAction],
        all_unitary_actions: List[BaseAction],
        all_unitary_actions_uncommon: List[BaseAction],
        sample_size: int = 1000,
        most_common_station: int = 16,
        show_progress: bool = False,
) -> List[Tuple[float, EncodedTopologyAction]]:
    """Search for good actions that occur on one or two substations at the same time.

    Note: In the l2rpn_neurips_2020_track1_small station 16 is heavily dominating as a station that is affected
    by unitary actions, hence we made a special case for single stations here.

    Args:
        env: The environment to use for the search.
        best_unitary_actions: List of the best unitary actions, used to find tuple actions.
        all_unitary_actions: List of all unitary actions.
        all_unitary_actions_uncommon: The list of all actions not affecting the most
                                      common station with id of most_common_station.
        sample_size: How many other actions should be sampled from all actions to combine with best actions.
        most_common_station: The id of the most common station in the network/environment.
        show_progress: Set to true to show a progress bar.

    Returns:
        A list of tuple actions ranked by their rho_improvement.
    """
    obs: BaseObservation = env.get_obs()
    min_rho, overflow_id = obs.rho.max(), obs.rho.argmax()
    old_max_rho = min_rho  # save old max_rho for comparison
    actions_results: List[Tuple[float, EncodedTopologyAction]] = []

    # Unitary search
    # For each of these actions simulate the unitary effect and write out rho_max.
    for action in tqdm(best_unitary_actions, disable=not show_progress):
        valid, reason = env._game_rules(action, env)
        if not valid:
            logging.debug(f"Found action is not valid: {reason}")
            continue

        obs_, _, done, _ = obs.simulate(action)
        if not done:
            new_max_rho = obs_.rho.max()
            rho_improvement: float = old_max_rho - new_max_rho
            actions_results.append((rho_improvement, EncodedTopologyAction(action)))

    # Tuple search
    for first_action in tqdm(best_unitary_actions, disable=not show_progress):
        first_affected_substations = affected_substations(first_action)
        assert len(first_affected_substations) == 1, "first action is unitary"
        first_affected_substation = first_affected_substations[0]

        affected_substation_id = int(first_action.as_dict()["set_bus_vect"]["modif_subs_id"][0])
        if affected_substation_id == most_common_station:
            # (2) if the first action is from substation 16 then loop through all other actions.
            for second_action in all_unitary_actions_uncommon:
                tuple_action = first_action + second_action
                valid, reason = env._game_rules(action, env)
                if not valid:
                    logging.debug(f"Found action is not valid: {reason}")
                    continue
                affected_substation = affected_substations(second_action)[0]
                if first_affected_substation == affected_substation:
                    continue
                assert len(affected_substations(tuple_action)) == 2
                obs_, _, done, info = obs.simulate(tuple_action)
                if not done:
                    new_max_rho = obs_.rho.max()
                    rho_improvement: float = old_max_rho - new_max_rho
                    if rho_improvement > 0:
                        logging.debug(f"[1] Found good tuple action {action} with {rho_improvement}")
                        actions_results.append((rho_improvement, EncodedTopologyAction(tuple_action)))
                    else:
                        logging.debug(f"Tuple action has bad rho_improvement: {rho_improvement}")
        else:
            # (1) if the first action is not from substation 16 then sample other actions with sample size e.g. 1000;
            # without replacement and check that actions are legal;
            random_indices = np.random.choice(len(all_unitary_actions), sample_size, replace=False)
            logging.debug(f"randomly sample other actions for {affected_substation_id}")
            for ind in random_indices:
                second_action = all_unitary_actions[ind]
                tuple_action = first_action + second_action
                valid, reason = env._game_rules(tuple_action, env)
                if not valid:
                    logging.debug(f"Action is invalid: reason: {reason}")
                    continue
                affected_substation = affected_substations(second_action)[0]
                if first_affected_substation == affected_substation:
                    continue
                assert len(affected_substations(tuple_action)) == 2
                obs_, _, done, info = obs.simulate(tuple_action)
                if not done:
                    new_max_rho = obs_.rho.max()
                    rho_improvement: float = old_max_rho - new_max_rho
                    if rho_improvement > 0:
                        logging.debug(f"[2] Found good tuple action {action} with {rho_improvement}")
                        actions_results.append((rho_improvement, EncodedTopologyAction(tuple_action)))

    best_actions = sorted(actions_results, key=lambda tup: tup[0], reverse=True)
    return best_actions


def topology_search_triples(
        env: BaseEnv,
        best_tuple_actions: List[BaseAction],
        all_unitary_actions: List[BaseAction],
        sample_size: int = 1000,
        show_progress: bool = False,
) -> List[Tuple[float, EncodedTopologyAction]]:
    """Search for good actions that occur on three substations at the same time, by combining unitary and tuple actions.

    Args:
        env: The environment used for the search.
        best_tuple_actions: The best tuple actions, found by topology_search_tuples.
        all_unitary_actions: All unitary actions of the environment.
        sample_size: How many unitary actions should be tried to be combined with tuple actions.
        show_progress: Set to true to show a progress bar.

    Returns:
        A list of triple actions ranked by their rho_improvement.
    """
    obs: BaseObservation = env.get_obs()
    min_rho, overflow_id = obs.rho.max(), obs.rho.argmax()
    old_max_rho = min_rho  # save old max_rho for comparison
    n_unitary_actions = len(all_unitary_actions)
    actions_results: List[Tuple[float, EncodedTopologyAction]] = []

    # Triple search
    for first_action in tqdm(best_tuple_actions, disable=not show_progress):
        first_affected_substations = affected_substations(first_action)
        assert len(first_affected_substations) == 2, "first action is tuple action"

        # Sample actions
        actions_to_sample = sample_size
        while actions_to_sample > 0:
            second_action = all_unitary_actions[randint(0, n_unitary_actions - 1)]
            if affected_substations(second_action)[0] in first_affected_substations:
                # Resample if substation already affected in first_action
                continue
            triple_action = first_action + second_action
            assert len(affected_substations(triple_action)) == 3
            obs_, _, done, info = obs.simulate(triple_action)
            if not done:
                new_max_rho = obs_.rho.max()
                rho_improvement: float = old_max_rho - new_max_rho
                if rho_improvement > 0:
                    logging.debug(f"[2] Found good triple action {triple_action} with {rho_improvement}")
                    actions_results.append((rho_improvement, EncodedTopologyAction(triple_action)))

            actions_to_sample -= 1

    best_actions = sorted(actions_results, key=lambda tup: tup[0], reverse=True)
    return best_actions


def topology_search_sequential_x_steps(
        env: BaseEnv, sub_action_set: dict, steps: int = 1, show_progress: bool = False
) -> List[Tuple[float, BaseAction]]:
    """Specific topology search that creates tuple, triple or more sequential action sets based on
    a provided set of sub-actions. This method does not require to overwrite the number of possible
    actions, because it ensures that for each step the action is executed.

    Note that this Agent takes the number of steps in the environment as provided.

    Args:
        env: Grid2Op Environment,
        sub_action_set: Dictionary containing all actions of the action set, seperated by substation,
        steps: Number of steps,
        show_progress: Set to true to show a progress bar,

    Returns:
        Action set.

    """

    obs: BaseObservation = env.get_obs()
    min_rho, overflow_id = obs.rho.max(), obs.rho.argmax()

    actions_results: List[Tuple[float, EncodedTopologyAction]] = []

    full_action = env.action_space({})
    performed_actions = []
    effected_substations = []
    rho_improvement = [min_rho]
    done = False
    # a[0].as_dict()['set_bus_vect']['modif_subs_id'][0]==
    for step in range(steps):
        best_id = None
        best_action = env.action_space({})
        if done:
            break

        new_max_rho = min_rho

        # Select all actions that were not effected:
        available_actions = []
        for k, v in sub_action_set.items():
            if k not in effected_substations:
                available_actions += v

        # For each step we run through the whole set of sub_actions and only select the best action
        for idx, action in tqdm(enumerate(available_actions), disable=not show_progress):
            valid, reason = env._game_rules(action, env)  # noqa
            if not valid:
                logging.debug(f"Found action is not valid: {reason}")
                continue
            if action.as_dict()["set_bus_vect"]["modif_subs_id"][0] in effected_substations:
                logging.debug("Substation of action already in action set")
                continue

            obs_, _, ddone, info = obs.simulate(action)

            valid = is_valid(observation=obs, act=action, done_sim=ddone, info_sim=info)

            if valid and (obs_.rho.max() < new_max_rho):
                new_max_rho = obs_.rho.max()
                best_action = action
                best_id = idx

        # Save best action:
        if best_id:
            performed_actions.append(best_id)
            effected_substations.append(best_action.as_dict()["set_bus_vect"]["modif_subs_id"][0])

        full_action = full_action + best_action

        # Based on the best action, we now run one step in the environment:
        # Note that we reconnect if possible. This might dilute the overall performacne.
        best_action = find_best_line_to_reconnect(obs, best_action)
        obs, _, done, _ = env.step(best_action)
        min_rho = obs.rho.max()
        rho_improvement.append(min_rho)

    # print results:
    logging.info(
        f"The following action ids where selected {performed_actions} which affected the substations"
        f"{effected_substations} and led to a reduction of {rho_improvement}."
    )

    return full_action



# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
import logging
import os
import shutil
import warnings
from pathlib import Path
from typing import Union, List, Optional, Tuple

import grid2op
import numpy as np
import ray
from grid2op.Agent import BaseAgent
from lightsim2grid import LightSimBackend

from curriculumagent.junior.junior_student import train as train_junior
from curriculumagent.senior.senior_student import Senior
from curriculumagent.submission.my_agent import MyAgent
from curriculumagent.teacher.collect_teacher_experience import make_unitary_actionspace
from curriculumagent.teacher.teacher import general_teacher
from curriculumagent.tutor.collect_tutor_experience import generate_tutor_experience, prepare_dataset


class CurriculumAgent(BaseAgent):
    """
    This is the Baseline of the Curriculum Agent, which in turn is a refurbished agent based on the submission of
    @https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution. Their work was published under a Mozilla Public
    Licence, so in turn, please credit their work as well, when using this code.

    The overall agent follows the typical baseline format. However, given that the original code consists of four
    different sub-agents, we decided to separate the agents in order to make it easier to understand.
    With the full pipeline, the CurriculumAgent consist of a Teacher,a Tutor, a Junior and a Senior agent that
    all contribute to the final MyAgent. In order to train the full pipeline, please execute the train_full_pipeline()
    method. Given that four different agents have to be trained,
    and the teacher + tutor agents are based on simply gready searches, the overall training might take some time.

    However, if you just want to use the agent or retrain the Senior (RL-Model), you can use the train method.
    The default training results were created for the IEEE14 of the grid2op package, named "l2rpn_case14_sandbox".

    """

    def __init__(self,
                 action_space: grid2op.Action.ActionSpace,
                 observation_space: grid2op.Observation.BaseObservation,
                 name: str,
                 **kwargs):
        """ Init Method of the CurriculumAgent. Note that the agent has to be loaded after initilization in order to
        work.

        Args:
            action_space: Action_space of Grid2Op Environment
            observation_space: Observation Space of the Grid2Op Environment
            name: Name of the Agent

            **kwargs: Additional values that might be helpfull when loading the MyAgent
        """
        BaseAgent.__init__(self, action_space)
        self.name = name
        self.observation_space = observation_space
        self.agent: MyAgent = None
        self.ckpt_path = None
        self.do_nothing = self.action_space({})
        self.name = name
        self.senior: Senior = None

        self.__kwargs = {}
        if any(kwargs.keys()):
            self.__kwargs = kwargs

    def act(self, observation: grid2op.Observation.BaseObservation,
            reward: float, done: bool) -> grid2op.Action.BaseAction:
        """ This is the act function of the CurriculumAgent.

        Args:
            observation: Grid2Op Observation
            reward: Reward of the previous step
            done: Whether the environment is done.

        Returns: Action of the agent

        """
        if self.agent is None:
            warnings.warn("The agent was not yet loaded, thus we return the do-nothing action. "
                          "Please run the load() for the CurriculumAgent.")
            action = self.do_nothing
        else:
            action = self.agent.act(observation=observation, reward=reward, done=done)
        return action

    def reset(self, observation: grid2op.Observation.BaseObservation) -> None:
        """
        This method is called at the beginning of a new episode.
        It is implemented by baselines to reset their internal state if needed.
        This method is not required for the CurriculumAgent.

        Args:
            observation: Observation of the Grid2Op Environment

        Returns: None

        """
        self.agent.reset(observation)

    def load(self, path: Union[str, Path],
             actions_path: Optional[Union[Path, List[Path]]] = None,
             **kwargs) -> None:
        """This method loads the MyAgent and inits the agent.

        We provide three ways to load the agent:
        1. You provide the Senior model path (preferred) of a trained Senior model
        2. You can provide a Junior model path, prior to the training.


        Note:
            We load with the default values, if you want to pass specific values to the MyAgent method,
            you can do that via kwargs. These can be for example the boolean subset (whether you
            do not want the whole observation) or the boolean topo (whether you want to revert your topology).
            Depending on the underlying model, you might need to train with the kwarg: {"subset":True}

        Args:
            path: path, where we can find the model for execution
            actions_path: path, where to find the action sets. This is required to run the agent. If this is not
            provided, we try to load from the same directory as the model.
            kwargs: You can pass further arguments to the MyAgent

        Returns:
            None.

        """
        # Check both input paths:
        action_path, model_path = self.__check_paths(path,
                                                     actions_path)

        if any(kwargs.keys()):
            for k, v in kwargs.items():
                self.__kwargs[k] = v

        self.agent = MyAgent(action_space=self.action_space,
                             model_path=model_path,
                             action_space_path=action_path,
                             **self.__kwargs)

    def save(self, path: Union[str, Path]) -> None:
        """ This method saves the model of the agent. Note that this requires the agent to
        be initialized

        Args:
            path: Where to save the model

        Returns: None

        """
        if self.agent is None:
            raise ValueError("Model can not be saved, because there is no agent. Please initialize the agent.")

        # Create paths:

        model_path = Path(path) / "model"
        if not model_path.is_dir():
            os.mkdir(model_path)
        self.agent.model.save(model_path)
        logging.info(f"Model is saved under {model_path}")

        action_path = Path(path) / "actions"
        if not action_path.is_dir():
            os.mkdir(action_path)
        np.save(action_path / "actions.npy", self.agent.actions)
        logging.info(f"Actions are saved under {action_path / 'actions.npy'}")

    def train(self, env: Union[grid2op.Environment.BaseEnv, str] = "l2rpn_case14_sandbox",
              iterations: int = 1,
              save_path: Union[Path, str] = Path(os.getcwd()) / "out",
              **kwargs_senior):
        """
        For the training, we initialize the Senior student and then run the training

        Note that in order for this to work, we need Ray. If it is not initialized, we
        call ray.init(). Accordingly, if you have specific cluster resources, make sure to
        call ray.init() prior to training this agent!

        Args:
            env: Grid2Op Environment
            iterations: Number of iterations to run the senior
            save_path: Where to save the model and the rllib checkpoints.
            **kwargs_senior: Additional arguments for the Senior model

        Returns: MyAgent

        """
        if isinstance(env, grid2op.Environment.BaseEnv):
            env_path = env.get_path_env()
        else:
            env_path = env

        # First we save the current model to ensure that we have the correct actions:
        assert self.agent is not None, "The agent has not been loaded. Please load the agent first!"
        self.save(save_path)

        # Test if files are there:
        action_path, model_path = self.__check_paths(save_path)

        # Gather all variables:
        ckpt_path = save_path / "ckeckpoints"
        if not ckpt_path.is_dir():
            os.mkdir(ckpt_path)
        logging.info("Saving previous model, action and checkpoint directory done!")
        # Ray:
        if not ray.is_initialized:
            ray.init()

        # Train the senior
        self.senior = Senior(env_path=env_path,
                             action_space_path=action_path,
                             model_path=model_path,
                             ckpt_save_path=ckpt_path,
                             **kwargs_senior)

        logging.info(f"Initiation of Senior done, let's start training with {iterations}.")
        self.senior.train(iterations=iterations)
        logging.info(f"Training complete. Overwriting agent")
        self.agent = self.senior.get_my_agent(path=model_path)
        logging.info(f"Completion of training")

    def train_full_pipeline(
            self,
            env: Union[str, grid2op.Environment.BaseEnv] = "l2rpn_case14_sandbox",
            iterations: int = 1,
            save_path: Optional[Path] = Path(os.getcwd()) / "out",
            **kwargs,
    ) -> grid2op.Agent.BaseAgent:
        """
        This method is used to train the full pipeline of the Curriculum Agent,i.e., the Teacher-Tutor-Junio-Senior
        framework.

        This means, that after initializing all directories, the Teacher will search for suitable
        actions of the Grid2Op environment. Afterwards, the Tutor will take a subset of these actions
        and generate an action-observation experience set. Then, the Junior (feed-forward) model, will
        learn to imitat the Tutor. As a last step the Senior receives the weights of the junior and
        learns with Reinforcement Learning to navigate in the environment.

        Note, it is necessary to provide a suiting Grid2Op environment. Further, expect that this might
        take a while.

        Args:
            env: Str, Path to environment or the grid2op Environment.
            iterations: Number of iterations for the training
            save_path: Optional save_path argument, if you want to store your agent somewhere else.
            **kwargs: Optional Parameters such as seed or jobs, or max_action_space_size

        Returns: Returns the fitted agent.

        """
        # Check for kwargs:
        seed = kwargs.get("seed", 42)
        jobs = kwargs.get("jobs", os.cpu_count())
        max_actionspace_size = kwargs.get("max_actionspace_size", 250)
        # Set seed:
        np.random.seed(seed)

        # Prepare environment
        log_format = "(%(asctime)s) [%(name)-10s] %(levelname)8s: %(message)s [%(filename)s:%(lineno)s]"
        logging.basicConfig(level=logging.INFO, format=log_format)

        # Make sure environment is present or downloaded.
        # For parallelization, this needs to be only the name
        if isinstance(env, grid2op.Environment.BaseEnv):
            env_path = env.get_path_env()
        else:
            env_path = env

        try:
            grid2op.make(env_path, backend=LightSimBackend())
        except Exception as e:
            raise ValueError(f"Could not init the Grid2op Enviornment with LightSimBackend. Please ensure that" \
                             f"this works! We raise a {e}")

        for sub_dir in ["teacher", "tutor", "junior", "senior", "model", "actions"]:
            (save_path / sub_dir).mkdir(exist_ok=True, parents=True)

        logging.info(f"All paths were created under {save_path}. Now let's start with the Teacher training.")

        ############################################################################################
        # Stage1 : Teacher
        # Run Teacher / Do Action Space Reduction
        teacher_experience_path = save_path / "teacher" / Path("general_teacher_experience.csv")
        if not teacher_experience_path.exists():
            logging.info(f"Finding good actions using Teacher...")
            general_teacher(
                save_path=teacher_experience_path,
                env_name_path=env_path,
                n_episodes=iterations,
                jobs=jobs,
            )
        else:
            logging.info(f"Skipping Teacher because {teacher_experience_path} already exists")

        # Use experience to do action space reduction and save as our action space file
        actions_path = save_path / "actions" / "actions.npy"
        if not actions_path.exists():
            logging.info("Reducing actionspace...")
            make_unitary_actionspace(
                actions_path, [teacher_experience_path], env_path,
                best_n=max_actionspace_size
            )
        else:
            logging.info(f"Skipping action space generation because {actions_path} already exists")

        ############################################################################################
        # Stage 2 : Tutor
        tutor_experience_path = save_path / "tutor" / "tutor_experience.npy"
        if not tutor_experience_path.exists():
            logging.info("Generating experience using Tutor...")
            generate_tutor_experience(
                env_path,
                tutor_experience_path,
                actions_path,
                num_chronics=iterations,
                jobs=jobs,
                seed=seed,
            )
        else:
            logging.info(f"Skipping Tutor because {tutor_experience_path} already exists")

        ############################################################################################
        # Stage 3: Junior
        junior_data_path = save_path / "tutor" / "junior_data"
        junior_results_path = save_path / "junior"
        junior_data_path.mkdir(exist_ok=True, parents=True)
        prepare_dataset(
            traindata_path=tutor_experience_path.parent,
            target_path=junior_data_path,
            dataset_name="test",
        )

        if not (junior_results_path / "saved_model.pb").exists():
            train_junior(
                run_name="junior",
                dataset_path=junior_data_path,
                target_model_path=junior_results_path,
                action_space_file=actions_path,
                dataset_name="test",
                epochs=10 + iterations,
                seed=seed,
            )
        else:
            logging.info(f"Skipping Junior because {junior_results_path / 'saved_model.pb'} already exists")

        ############################################################################################
        # Stage 4: Senior:
        # Configure configs & register model:
        # Given some import problems we only import ray here:
        senior_results_path = save_path / "senior"
        if ray.is_initialized() is False:
            ray.init()

        # get the resources:
        resources = ray.nodes()
        num_workers = int(resources[0]["Resources"]["CPU"] // 2)

        senior = Senior(env_path=env_path,
                        action_space_path=actions_path,
                        model_path=junior_results_path,
                        ckpt_save_path=senior_results_path,
                        num_workers=num_workers,
                        subset=False)

        if not (senior_results_path / "sandbox").exists():
            senior.train(iterations=iterations)
        else:
            logging.info(f"Skipping Senior training because {senior_results_path / 'sandbox'} already "
                         f"exists. Instead we are loading the senior from checkpoints")  #
            senior.restore(senior_results_path)
            logging.info(f"The Senior trained for {senior.ppo.iteration} iterations.")

        ############################################################################################
        # Stage 5.
        # Saving the agent

        agent_path = save_path / "model"
        logging.info(f"Saving the agent model and building the agent")
        self.agent = senior.get_my_agent(path=agent_path)
        logging.info(f"Loading of Agent Succeeded. Return MyAgent.")
        ray.shutdown()
        return self.agent

    def __check_paths(self, path: Union[str, Path],
                      action_path: Optional[Union[str, Path]] = None) -> Tuple[Path, Path]:
        """ Check whether the path is in the required format

        Args:
            path: Path, where the agent should be saved
            action_path: Optional path for the actions, if they were specified differently.

        Returns:

        """
        path = Path(path)
        assert path.is_dir(), "The provided path does not exist. Please pass an existing directory."

        # Action Paths:
        if action_path is None:
            action_path = path / "actions"

        if not isinstance(action_path, list):
            if not action_path.is_dir():
                warnings.warn(f"We could not find the directory{action_path}. We try to import directly from {path}.")
                action_path = path

            if not any([".npy" in a_p for a_p in os.listdir(action_path)]):
                raise FileNotFoundError(f"No .npy was found in the action path {action_path}. Thus, we "
                                        f"can not load the actions")
        else:
            if not any([".npy" in str(a_p) for a_p in action_path]):
                raise FileNotFoundError(f"No .npy was found in the action path list {action_path}. Thus, we "
                                        f"can not load the actions")

        # Model Paths
        model_path = path / "model"
        if not model_path.is_dir():
            warnings.warn(f"We could not find the directory{model_path}. We try to import directly from {path}.")
            model_path = path

        if not any(["saved_model.pb" in a_p for a_p in os.listdir(model_path)]):
            raise FileNotFoundError(f"We could not find any tensorflow model in the directory {model_path}. "
                                    f"Please provide the correct path.")

        return action_path, model_path

    def create_submission(self, path: Union[str, Path] = "./submission"):
        """ Method that takes a saved model and agent and creates a full submission folder in the provided path.

        This only covers the "basic" CurriculumAgent!

        Note that you use a scaler, you have to change the make agent method and add the path. Same holds
        for other configurations.

        Args:
            path: Path where to save the submission

        Returns:
            None

        """
        assert path.is_dir(), "The provided path does not exist. Please pass an existing directory."

        # Save the model in the path:
        self.save(Path(path))

        # Creating directories
        path_for_files = Path(__file__).parent.parent
        path_for_common = Path(path) / "common"

        if not path_for_common.is_dir():
            os.mkdir(path_for_common)

        # Now copying the files from the commons method:
        shutil.copy(path_for_files / "common" / "__init__.py", path_for_common)
        shutil.copy(path_for_files / "common" / "obs_converter.py", path_for_common)
        shutil.copy(path_for_files / "common" / "utilities.py", path_for_common)

        # Move agent
        shutil.copy(path_for_files / "submission" / "my_agent.py", path)
        shutil.copy(path_for_files / "submission" / "__init__.py", path)

        logging.info(f"Moved required files to {path} .")
        logging.info("Note, if you have specific settings (scaler, subset, ...) you have to manually change the "
                     "my_agent method.")



__all__ = [
    "CurriculumAgent",
    "evaluate",
    "train"
]

from curriculumagent.baseline.baseline import CurriculumAgent
from curriculumagent.baseline.evaluate import evaluate
from curriculumagent.baseline.train import train



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
from pathlib import Path
from typing import Union, Optional

import grid2op
import ray

from curriculumagent.baseline import CurriculumAgent


def train(
        env: grid2op.Environment.BaseEnv,
        name: str = "Training_Pipeline",
        iterations: int = 1,
        save_path=Optional[Union[Path, str]],
        **kwargs
):
    """
    This is the method to train the full pipeline of the CurriculumAgent.
    For a quick training the CurriculumAgent we recommend to use the agent.train() method to only
    train the RL agent. This, however, requires that you already have an model and an action set.

    If you want/have to start at zero and need to find actions as well as the model, feel free to run
    this pipeline. This pipeline will create for each step (agent) in the pipeline a directory.

    Note two things:
    1. This pipeline does not require a loading of the CurriculumAgent and thus does not need an action
    set and a pretrained model. You start from scratch.
    2. This execution is computationally expensive. Further, your machine needs to be able to run RLlib.

    Args:
        env: Grid2op Environment.
        name: Name of the Training agent.
        iterations: Number of iterations to train.
        save_path: Optional Save path where to save the model.
        **kwargs: Additional arguments you want to pass to the train_full_pipeline.

    Returns:
        None.

    """
    baseline = CurriculumAgent(action_space=env.action_space, observation_space=env.observation_space, name=name)

    baseline.train_full_pipeline(env=env, name=name, iterations=iterations, save_path=save_path,**kwargs)
    ray.shutdown()



if __name__ == "__main__":
    """
    This is a possible implementation of the train script.
    """
    import grid2op
    from l2rpn_baselines.utils import cli_train

    args_cli = cli_train().parse_args()
    env = grid2op.make()
    train(
        env=env,
        name=args_cli.name,
        iterations=args_cli.num_train_steps,
        save_path=args_cli.save_path,
    )



"""This file consist the CreateJuniorData class which converts the tutor data into readable data for the
junior agent.

"""
import logging
import os
import random
import time
from datetime import datetime
from multiprocessing import Pool
from pathlib import Path
from typing import List, Union, Tuple, Optional

import defopt
import grid2op
import numpy as np
from grid2op.Agent import BaseAgent
from grid2op.Environment import BaseEnv

from curriculumagent.common.obs_converter import obs_to_vect
from curriculumagent.common.utilities import split_and_execute_action, find_best_line_to_reconnect
from curriculumagent.tutor.tutors.general_tutor import GeneralTutor


def collect_tutor_experience_one_chronic(
        action_paths: Union[Path, List[Path]],
        chronics_id: int,
        env_name_path: Union[Path, str] = "l2rpn_neurips_2020_track1_small",
        seed: Optional[int] = None,
        enable_logging: bool = True,
        subset: Optional[Union[bool, str]] = False,
        TutorAgent: BaseAgent = GeneralTutor,
        tutor_kwargs: Optional[dict] = {},
        env_kwargs: Optional[dict] = {}
):
    """Collect tutor experience of one chronic.

    Args:
        action_paths: List of Paths for the tutor.
        chronics_id: Number of chronic to run.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        seed: Whether to init the Grid2Op env with a seed
        enable_logging: Whether to log the Tutor experience search.
        subset: Optional argument, whether the observations should be filtered when saved.
            The default version saves the observations according to obs.to_vect(), however if
            subset is set to True, then only the all observations regarding the lines, busses, generators and loads are
            selected. Further note, that it is possible to say "graph" in order to get the connectivity_matrix as well.
        TutorAgent: Tutor Agent which should be used for the search.
        tutor_kwargs: Additional arguments for the tutor, e.g. the max rho or the topology argument.
        env_kwargs: Optional arguments that should be used when initializing the environment.

    Returns:
        None.
    """
    if enable_logging:
        logging.basicConfig(level=logging.INFO)

    try:
        # if lightsim2grid is available, use it.
        from lightsim2grid import LightSimBackend

        backend = LightSimBackend()
        env = grid2op.make(dataset=env_name_path, backend=backend,**env_kwargs)
    except ImportError:  # noqa
        env = grid2op.make(dataset=env_name_path,**env_kwargs)
        logging.warning("Not using lightsim2grid! Operation will be slow!")

    if seed:
        env.seed(seed)

    env.set_id(chronics_id)
    env.reset()
    logging.info(f"current chronic:{env.chronics_handler.get_name()}")

    # After initializing the environment, let's init the tutor
    if tutor_kwargs is None:
        tutor_kwargs = {}
    else:
        logging.info(f"Run Tutor with these additional kwargs {tutor_kwargs}")

    tutor = TutorAgent(action_space=env.action_space, action_space_file=action_paths, **tutor_kwargs)
    # first col for label which is the action index, remaining cols for feature (observation.to_vect())
    done, step, obs = False, 0, env.get_obs()
    if subset:
        vect_obs = obs_to_vect(obs)
    elif subset == "graph":
        vect_obs = obs_to_vect(obs, True)
    else:
        vect_obs = obs.to_vect()

    records = np.zeros((1, 1 + len(vect_obs)), dtype=np.float32)

    while not done:
        action, idx = tutor.act_with_id(obs)
        if action.any() and (idx != -1):
            # Note that we exclude the TOPO Actions of the Tutor!!!
            if subset:
                vect_obs = obs_to_vect(obs)
            elif subset == "graph":
                vect_obs = obs_to_vect(obs, True)
            else:
                vect_obs = obs.to_vect()

            records = np.concatenate(
                (records, np.hstack([idx, vect_obs]).astype(np.float32).reshape(1, -1)), axis=0
            )

            # Execute Action:
            # This method does up to three steps and returns the output
            obs, _, done, _ = split_and_execute_action(env=env, action_vect=action)
            step = env.nb_time_step
            logging.info(f"game over at step-{step}")

        else:
            # Use Do-Nothing Action
            act_with_line = find_best_line_to_reconnect(obs, env.action_space({}))
            obs, _, done, _ = env.step(act_with_line)

    return records

    # save current records


def generate_tutor_experience(
        env_name_path: Union[Path, str],
        save_path: Union[Path, str],
        action_paths: Union[Path, List[Path]],
        num_chronics: Optional[int] = None,
        num_sample: Optional[int] = None,
        jobs: int = -1,
        subset: Optional[Union[bool, str]] = False,
        seed: Optional[int] = None,
        TutorAgent: BaseAgent = GeneralTutor,
        tutor_kwargs: Optional[dict] = {},
        env_kwargs: Optional[dict] = {},
):
    """Method to run the Tutor in parallel.

    Args:
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        save_path: Where to save the experience.
        action_paths: List of action sets (in .npy format).
        num_chronics: Total numer of chronics.
        num_sample: Length of sample from the num_chronics. If num_sample is smaller than num chronics,
            a subset is taken. If it is larger, the chronics are sampled with replacement.
        subset: Optional argument, whether the observations should be filtered when saved.
            The default version saves the observations according to obs.to_vect(), however if
            subset is set to True, then only the all observations regarding the lines, busses, generators and loads are
            selected. Further, note that it is possible to say graph in order to get the connectivity_matrix as well.
        jobs: Number of jobs in parallel.
        seed: Whether to set a seed to the sampling of environments
        TutorAgent: Tutor Agent which should be used for the search, default is the GeneralTutor.
        tutor_kwargs: Optional arguments that should be passed to the tutor agent.
        env_kwargs: Optional arguments that should be used when initializing the environment.

    Returns:
        None, saves results as numpy file.

    """
    log_format = "(%(asctime)s) [%(name)-10s] %(levelname)8s: %(message)s [%(filename)s:%(lineno)s]"
    logging.basicConfig(level=logging.INFO, format=log_format)

    if jobs == -1:
        jobs = os.cpu_count()

    tasks = []

    # Make sure we can initialize the environment
    # This also makes sure that the environment actually exits or gets downloaded
    env: BaseEnv = grid2op.make(env_name_path)
    chronics_path = env.chronics_handler.path
    if chronics_path is None:
        raise ValueError(f"Can't determine chronics path of given environment {env_name_path}")

    if num_chronics is None:
        num_chronics = len(os.listdir(chronics_path))

    if num_sample:
        if num_sample <= num_chronics:
            sampled_chronics = random.sample(range(num_chronics), num_sample)
        else:
            sampled_chronics = random.choices(np.arange(num_chronics), k=num_sample)
    else:
        sampled_chronics = np.arange(num_chronics)

    for chronic_id in sampled_chronics:
        tasks.append((action_paths, chronic_id, env_name_path, seed, True, subset, TutorAgent, tutor_kwargs,env_kwargs))
    if jobs == 1:
        # This makes debugging easier since we don't fork into multiple processes
        logging.info(f"The following {len(tasks)} tasks will executed sequentially: {tasks}")
        out_result = []
        for task in tasks:
            out_result.append(collect_tutor_experience_one_chronic(*task))
    else:
        logging.info(f"The following {len(tasks)} tasks will be distributed to a pool of {jobs} workers:")
        start = time.time()
        with Pool(jobs) as p:
            out_result = p.starmap(collect_tutor_experience_one_chronic, tasks)
        end = time.time()
        elapsed = end - start
        logging.info(f"Time: {elapsed}s")

    # Now concatenate the result:
    all_experience = np.concatenate(out_result, axis=0)
    if save_path.is_dir():
        now = datetime.now().strftime("%d%m%Y_%H%M%S")
        save_path = save_path / f"tutor_experience_{now}.npy"

    np.save(save_path, all_experience)
    logging.info(f"Tutor experience has been saved to {save_path}")


def prepare_dataset(
        traindata_path: Path,
        target_path: Path,
        dataset_name: str,
        extend: bool = False,
        seed: Optional[int] = 42,
):
    """Prepare/process the training data given by the tutor. The data is seperated into
    training, validation and test dataset, saved at the target_path.

    Args:
        traindata_path: The path to the generated records_* files by the tutor to use.
        target_path: The path to save the dataset files to.
        dataset_name: Name of the dataset which gets used to create the files.
        extend: Whether to extend the existing training data instead of creating a new dataset.
        seed: The random seed for shuffling the samples.

    Returns:
        None.

    """
    np.random.seed(seed)

    assert traindata_path.exists() and traindata_path.is_dir(), f"{traindata_path} is not a valid directory"

    # Change this more globaly
    npy_files = list(traindata_path.glob("*.npy"))
    assert len(npy_files), f"There are no files at {traindata_path}"

    if not extend:
        logging.info(f"Creating final dataset at directory {target_path}")
        make_dataset(target_path=target_path, dataset_name=dataset_name, npy_files=npy_files)
        logging.info("Done")
    else:
        train_file_path = target_path / f"{dataset_name}_train.npz"
        extend_dataset(npy_files=npy_files, train_file_path=train_file_path)


def merge_dataset(npy_file_paths: List[Path]) -> np.ndarray:
    """Merge the given numpy files at npy_file_paths and return merged dataset.

    Note:
         If the input consists of tuple action IDs, then the IDs are merged to one id!

    Args:
        npy_file_paths:  Input numpy files, created by the Tutor.

    Returns:
        The unique actions in a merged array.

    """
    assert len(npy_file_paths) > 0, "No files given for merging"

    npy_file_paths = sorted(npy_file_paths)  # This will hopefully make this reproducible

    loaded_actions = [np.load(str(f)) for f in npy_file_paths]
    concated_actions = np.concatenate(loaded_actions, axis=0)
    unique_actions = np.unique(concated_actions, axis=0)

    logging.info(f"All Experience: {concated_actions.shape}")
    logging.info(f"Unique Experience: {unique_actions.shape}")
    logging.info(f"Unique Actions: {len(np.unique(concated_actions[:, 0], axis=0))}")
    return unique_actions


def create_dataset(
        target_path: Union[Path, str],
        data: np.ndarray,
        dataset_name: Optional[str] = "Junior",
        chosen: Optional[Union[bool, list]] = False,
):
    """Given all samples data, select the right features and labels and split the dataset up.
    Save the dataset in three different files at target_path.

    Args:
        target_path: Path, where to save the training, validation and testing data for the junior.
        data: Dataset for the junior model.
        dataset_name: Optional name for the data set.
        chosen: Boolean or List for filtering the observations of the original observations.

    Returns:
        None. Saves  '{dataset_name}_train.npz','{dataset_name}_val.npz','{dataset_name}_test.npz' in the
    target path.

    """
    # Shuffle actions
    np.random.shuffle(data)

    assert target_path.exists() and target_path.is_dir(), "Given target_path is not a directory that exists"

    # Subset
    if chosen:
        data = data[:, chosen]

    # Dataset partition
    num_sampling = data.shape[0]
    train_size = num_sampling * 8 // 10
    # validate_size = num_sampling // 10
    test_size = num_sampling // 10
    # s for state, feature; a for action, label.

    # Check whether the input is a Tuple ID or not. If yes
    s_train, a_train = data[:train_size, 1:], data[:train_size, :1].astype(int)
    s_validate, a_validate = data[train_size:-test_size, 1:], data[train_size:-test_size, :1].astype(int)
    s_test, a_test = data[-test_size:, 1:], data[-test_size:, :1].astype(int)

    logging.info(f"TrainSet Size: {len(s_train)}")
    logging.info(f"ValidationSet Size: {len(s_validate)}")
    logging.info(f"TestSet Size: {len(s_test)}")

    train_path = target_path / f"{dataset_name}_train.npz"
    np.savez_compressed(train_path, s_train=s_train, a_train=a_train)

    validation_path = target_path / f"{dataset_name}_val.npz"
    np.savez_compressed(validation_path, s_validate=s_validate, a_validate=a_validate)

    test_path = target_path / f"{dataset_name}_test.npz"
    np.savez_compressed(test_path, s_test=s_test, a_test=a_test)


def load_dataset(
        dataset_path: Union[str, Path], dataset_name: str
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Load the dataset for the training of the Junior model.

    Args:
        dataset_path: Path, where to find the dataset.
        dataset_name: Name of the dataset. There should be '{dataset_name}_train.npz',
        '{dataset_name}_val.npz' and a '{dataset_name}_test.npz' in the target path.


    Returns:
        Returns x,y values for train, val and test data.

    """
    train_path = dataset_path / f"{dataset_name}_train.npz"
    train_data = np.load(train_path)

    validation_path = dataset_path / f"{dataset_name}_val.npz"
    validation_data = np.load(validation_path)

    test_path = dataset_path / f"{dataset_name}_test.npz"
    test_data = np.load(test_path)

    return (
        train_data["s_train"],
        train_data["a_train"],
        validation_data["s_validate"],
        validation_data["a_validate"],
        test_data["s_test"],
        test_data["a_test"],
    )


def make_dataset(
        target_path: Path, dataset_name: str, npy_files: List[Path]):
    """Join all experience into one file and then execute the create_dataset.

    Args:
        target_path: Path, where to write the new experience.
        dataset_name: Name of the files.
        npy_files: list of np.ndarrays.

    Returns:
        None.

    """
    logging.info("Merging data...")
    experience = merge_dataset(npy_files)

    logging.info("Splitting data...")
    create_dataset(target_path=target_path, dataset_name=dataset_name, data=experience)


def extend_dataset(npy_files: List[Path], train_file_path: Path) -> object:
    """Extend existing Training Dataset with new files.

    Args:
        npy_files: List of np.ndarrays
        train_file_path: Path of the training data.

    Returns:
        An extended dataset.

    """
    logging.info(f"Trying to add data to existing dataset {train_file_path}")
    # 2. Add new data to train set

    # TODO: This is no proper caching!!!
    cache_file_path = Path("merged_new_data.npz")
    if cache_file_path.exists():
        new_experience = np.load(str(cache_file_path))["experience"]
    else:
        new_experience = merge_dataset(npy_files)

    final_dataset = np.load(str(train_file_path))
    s_train, a_train = final_dataset["s_train"], final_dataset["a_train"]

    np.random.shuffle(new_experience)

    s_train = np.concatenate([s_train, new_experience[:, 1:]])
    a_train = np.concatenate([a_train, new_experience[:, :1].astype(int)])

    new_filename = train_file_path.name.split(".")[0] + "_v2.npz"
    logging.info(f"Saving new dataset to {new_filename}")
    np.savez_compressed(new_filename, s_train=s_train, a_train=a_train)




"""This file contains the different teacher types of the CurriculumAgent. The attacking teacher and the
general teacher are based on the original submission of @https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution.

The remaining two teacher were created in the project AI2Go.
"""

import os
from pathlib import Path
from typing import Union, Optional

from curriculumagent.teacher.teachers.teacher_n_minus_1 import NMinusOneTeacher
from curriculumagent.teacher.teachers.teacher1 import run_attacking_teacher
from curriculumagent.teacher.teachers.teacher2 import run_general_teacher
from curriculumagent.teacher.teachers.tuple_triple_teacher import run_multi_teacher


def attacking_teacher(
        save_path: Path,
        env_name_path: Union[Path, str] = "l2rpn_neurips_2020_track1_small",
        *,
        n_episodes: int = 100,
        top_k: int = 125,
        jobs: int = os.cpu_count(),
        seed: Optional[int] = None
):
    """This teacher is trying to collect useful actions by attacking/disconnecting specific lines at random timesteps
    and trying to find the best action to resolve that problem.

    Args:
        save_path: Path to save the teacher_experience file to.
        env_name_path: Path to grid2op dataset or the standard name of it.
        n_episodes: Number of episodes(run through all scenarios) to run.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        jobs: Number of jobs to use for parallel teacher_experience collection.
        seed: Optional seed for the Grid2Op Environment and random sampling

    Returns:
        None.
    """

    run_attacking_teacher(
        save_path=save_path, env_name_path=env_name_path, n_episodes=n_episodes, top_k=top_k, jobs=jobs, seed=seed
    )


def general_teacher(
        save_path: Path,
        env_name_path: Union[Path, str] = "l2rpn_neurips_2020_track1_small",
        rho_acting_threshold: float = 0.925,
        *,
        n_episodes: int = 100,
        limit_chronics: int = None,
        top_k: int = 125,
        jobs: int = os.cpu_count(),
        seed: int = 42
):
    """This teacher is trying to collect useful actions by doing topology search while running an
    environment when an overflow occurs.

    Args:
        save_path: Path to save the teacher_experience file to.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        rho_acting_threshold: When the environments max_rho gues above this threshold the agent begins to act.
        n_episodes: Number of episodes to run (through all scenarios).
        limit_chronics: Limit the number of chronics to run through.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        jobs: Number of jobs to use for parallel teacher_experience collection.
        seed: The random seed to use.

    Returns:
        None.
    """
    run_general_teacher(
        save_path=save_path,
        env_name_path=env_name_path,
        rho_acting_threshold=rho_acting_threshold,
        n_episodes=n_episodes,
        limit_chronics=limit_chronics,
        top_k=top_k,
        jobs=jobs,
        seed=seed,
    )


def n_minus_1_teacher(
        save_path: Path,
        env_name_path: Union[Path, str] = "l2rpn_neurips_2020_track1_small",
        rho_n0_threshold: Optional[float] = 0.9,
        rho_max_threshold: Optional[float] = 1.0,
        jobs: int = -1,
        seed: int = 42,
):
    """Run N-1 Teacher to collect n-1 compatible action. In comparison to the other teachers,
    this teacher only runs through each chronic once. However, if an n-1 possibility occurs, the
    teacher will actively disconnect lines.

    Args:
        save_path: Path to save the teacher_experience file to.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        rho_n0_threshold: When to start searching for N-1 action.
        rho_max_threshold: When to stop searching for N-1 actions and using only greedy actions instead.
        jobs: Number of jobs, default is all available CPUs.
        seed: Optional seed to set the Grid2Op env with.

    Returns:
        None.
    """
    n1_agent = NMinusOneTeacher(rho_n0_threshold=rho_n0_threshold, rho_max_threshold=rho_max_threshold, seed=seed)
    n1_agent.collect_n_minus_1_experience(save_path, env_name_path, jobs)


def multi_teacher(
        save_path: Path,
        use_triple_search: bool,
        env_name_path: str = "l2rpn_neurips_2020_track1_small",
        rho_acting_threshold: float = 0.925,
        *,
        n_episodes: int = 100,
        top_k: int = 125,
        limit_chronics: int = None,
        jobs: int = os.cpu_count(),
        seed: int = 42
) -> Path:
    """This teacher is trying to collect useful actions by doing topology search while running an
    environment when an overflow occurs. It also tries to combine actions to form tuple or triple actions.

    Args:
        save_path: Path to save the teacher_experience file to.
        use_triple_search: Execute the triple teacher instead of the tuple teacher.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        rho_acting_threshold: When the environments max_rho goes above this threshold the agent begins to act.
        n_episodes: Number of episodes to run (through all scenarios).
        limit_chronics: Limit the number of chronics to run through.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        jobs: Number of jobs to use for parallel teacher_experience collection.
        seed: The random seed to use.

    Returns:
        None.
    """
    run_multi_teacher(
        save_path=save_path,
        use_triple_search=use_triple_search,
        env_name_path=env_name_path,
        rho_acting_threshold=rho_acting_threshold,
        n_episodes=n_episodes,
        top_k=top_k,
        limit_chronics=limit_chronics,
        jobs=jobs,
        seed=seed,
    )



#!/usr/bin/env python3

# Copyright (c) 2020, RTE (https://www.rte-france.com)
# See AUTHORS.txt
# This Source Code Form is subject to the terms of the Mozilla Public License, version 2.0.
# If a copy of the Mozilla Public License, version 2.0 was not distributed with this file,
# you can obtain one at http://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
# This file is part of L2RPN Baselines, L2RPN Baselines a repository to host baselines for l2rpn competitions.
import logging
from pathlib import Path
from typing import Union, Optional

import grid2op
from grid2op.Runner import Runner
from l2rpn_baselines.utils.save_log_gif import save_log_gif

from curriculumagent.baseline.baseline import CurriculumAgent


def evaluate(
        env: grid2op.Environment.BaseEnv,
        load_path: Union[str, Path] = ".",
        logs_path: Optional[Union[str, Path]] = None,
        nb_episode: int = 1,
        nb_process: int = 1,
        max_steps: int = -1,
        verbose: Union[bool, int] = False,
        save_gif: bool = False,
        **kwargs,
) -> Runner:
    """This is the evaluate method for the Curriculum Agent.

    Args:
        env: The environment on which the baseline will be evaluated. The default is the IEEE14 Case. For other
        environments please retrain the agent in advance.
        load_path: The path where the model is stored. This is used by the agent when calling "agent.load()"
        logs_path: The path where the agents results will be stored.
        nb_episode: Number of episodes to run for the assessment of the performance. By default, it equals 1.
        nb_process: Number of process to be used for the assessment of the performance. Should be an integer greater
        than 1. By default, it's equals 1.
        max_steps: Maximum number of timesteps each episode can last. It should be a positive integer or -1.
        -1 means that the entire episode is run (until the chronics is out of data or until a game over).
        By default,it equals -1.
        verbose: Verbosity of the output.
        save_gif:  Whether to save a gif into each episode folder corresponding to the representation of the said
        episode. Note, that depending on the environment (and the performance of your agent) this creation of the gif
        might take quite a lot of time!
        **kwargs:

    Returns:
        The experiment file consisting of the data.

    """
    runner_params = env.get_params_for_runner()
    runner_params["verbose"] = verbose

    # Create the agent (this piece of code can change)
    agent = CurriculumAgent(
        action_space=env.action_space,
        observation_space=env.observation_space,
        name="Evaluation"
    )
    # Load weights from file (for example)
    agent.load(load_path)

    # Build runner
    runner = Runner(**runner_params, agentClass=None, agentInstance=agent)

    # you can do stuff with your model here

    # start the runner

    if nb_process > 1:
        logging.warning(
            f"Parallel execution is not yet available for keras model. Therefore, the number of processes is comuted with "
            f"only one process."
        )
        nb_process = 1

    res = runner.run(path_save=logs_path, nb_episode=nb_episode, nb_process=nb_process, max_iter=max_steps, pbar=False)

    # Print summary
    logging.info("Evaluation summary:")
    for _, chron_name, cum_reward, nb_time_step, max_ts in res:
        msg_tmp = "\tFor chronics located at {}\n".format(chron_name)
        msg_tmp += "\t\t - cumulative reward: {:.6f}\n".format(cum_reward)
        msg_tmp += "\t\t - number of time steps completed: {:.0f} / {:.0f}".format(nb_time_step, max_ts)
        logging.info(msg_tmp)

    if save_gif:
        save_log_gif(logs_path, res)
    return res


if __name__ == "__main__":
    """
    This is a possible implementation of the eval script.
    """
    from lightsim2grid import LightSimBackend
    import grid2op

    logging.basicConfig(level=logging.INFO)
    env = grid2op.make("l2rpn_case14_sandbox", backend=LightSimBackend())
    obs = env.reset()
    path_of_model = Path(__file__).parent / "model_IEEE14"
    myagent = CurriculumAgent(
        action_space=env.action_space,
        observation_space=env.observation_space,
        model_path=path_of_model,
        path_to_data=path_of_model,
        name="Test",
    )
    env = grid2op.make("l2rpn_case14_sandbox")
    out = evaluate(
        env,
        load_path=path_of_model,
        logs_path=Path(__file__).parent / "logs",
        nb_episode=10,
        nb_process=1,
        max_steps=-1,
        verbose=0,
        save_gif=True,
    )



"""This file consist of the GeneralTutor class. In comparison to the original approach, this class offers
more variables to tweak the performance (as well as including multiple action spaces).

"""
import logging
import time
from pathlib import Path
from typing import Optional, Union, Tuple, List

import grid2op
import numpy as np
from grid2op.Action import BaseAction
from grid2op.Agent import BaseAgent
from grid2op.Observation import BaseObservation

from curriculumagent.common.utilities import simulate_action, find_best_line_to_reconnect, split_action_and_return, \
    map_actions


class GeneralTutor(BaseAgent):
    """The class of the tutor agent which takes a reduced action space acts greedily using it.
    """

    def __init__(
            self,
            action_space: grid2op.Action.ActionSpace,
            action_space_file: Union[Path, List[Path]],
            do_nothing_threshold: Optional[float] = 0.925,
            best_action_threshold: Optional[float] = 0.999,
            return_status: Optional[bool] = True,
    ):
        """Simplified __init__ method of the Tutor class.

        The required actions are either a Path variable leading to a numpy array or a list with
        multiple entries (paths). If multiple entries are supplied, the tutor executes the list sequentially.
        That means, if the threshold of the best action is met in the first set of actions from the
        list, the greedy search stops.

        Args:
            action_space: action space object from Gird2Op environment
            action_space_file: Either Numpy file with actions or List with multiple actions.
            do_nothing_threshold: Threshold, when the do nothing action is not sufficient.
            best_action_threshold: Threshold, when the collected action is sufficient and the search can
            be stopped.
            return_status: Whether each step should be logged.

        Returns:
            None.

        """
        super().__init__(action_space=action_space)

        # Set self actions to a list to iterate for later
        if isinstance(action_space_file, Path):
            assert action_space_file.is_file()
            list_of_actions = [np.load(str(Path(action_space_file)))]

        elif isinstance(action_space_file, list):
            for act_path in action_space_file:
                assert act_path.is_file()
            list_of_actions = [np.load(str(act_path)) for act_path in action_space_file]

        # Now map action ids:
        self.actions = map_actions(list_of_actions)

        self.action_space_size = self.action_space.size()
        self.do_nothing_threshold = do_nothing_threshold
        self.action_threshold = best_action_threshold
        self.return_status = return_status
        self.next_actions = None

    def act_with_id(self, observation: BaseObservation) -> Tuple[np.ndarray, int]:
        """Compute greedy search of Tutor.

        In this greedy search, we iterate over the keys of the dictionary.
        If the threshold is met, the iteration stops and the best action is returned.

        Args:
            observation: Grid2Op observation.

        Returns:
            Best action as numpy array as well as a tuple index for the action.

        """
        start_time = time.time()

        # Check for do nothing
        if observation.rho.max() < self.do_nothing_threshold:
            # secure, return "do nothing" in bus switches.
            return np.zeros(self.action_space_size), -1

        # Take lower values of either rho max or do nothing rho max
        obs_dn, _, _, _ = observation.simulate(self.action_space({}))

        min_rho = observation.rho.max()
        old_rho_max = min_rho
        action_chosen = None
        best_action_index = -1

        # Run through the list of actions
        # Set len act to 0 in order to assert the correct idx:

        for actions in self.actions:
            # actions is a dictionary with the actions
            for idx, action_array in actions.items():
                obs_sim, valid_action = simulate_action(
                    action_vect=action_array, action_space=self.action_space, obs=observation
                )
                if not valid_action:
                    continue

                # Simulate action (even though it might be illegal for tuple or triple action)
                if obs_sim < min_rho:
                    min_rho = obs_sim
                    action_chosen = action_array
                    best_action_index = idx

            if min_rho <= self.action_threshold:
                break

        if self.return_status:
            print_status(observation, best_action_index, old_rho_max, min_rho, start_time)

        if action_chosen is not None:
            out = action_chosen, best_action_index

        else:
            out = np.zeros(self.action_space_size), -1

        return out

    def act(self, observation: BaseObservation, reward: float, done: bool = False) -> BaseAction:
        """Compute greedy search of Tutor.

        This is wrapper for the act_with_id method.

        Args:
            observation: Grid2Op observation.

        Returns:
            Returns a base action.

        """
        next_action = find_best_line_to_reconnect(obs=observation, original_action=self.action_space({}))
        if self.next_actions is not None:
            # Try to do a step:
            try:
                next_action = next(self.next_actions)
            except StopIteration:
                self.next_actions = None

        if self.next_actions is None:
            act_array, _ = self.act_with_id(observation=observation)

            # Create generator
            self.next_actions = split_action_and_return(observation, self.action_space, act_array)
            next_action = next(self.next_actions)
            next_action = find_best_line_to_reconnect(obs=observation, original_action=next_action)

        return next_action


def print_status(observation, best_action_index: int, old_rho_max, min_rho, start_time) -> None:
    """Logging values for evaluation.

    Args:
        observation: Observation of Grid2Op.
        best_action_index: Combined index of the action.
        old_rho_max: Previous rho max that triggered the tutor.
        min_rho: The best rho value from the actions.
        start_time: Time variable for printing.

    Returns:
        None.

    """
    rho_improvement = old_rho_max - min_rho
    if best_action_index != -1:
        logging.info(
            f"t={observation.get_time_stamp()}, line={observation.rho.argmax():03d} overflowed"
            f"=> Use action {best_action_index} => rho_delta: {rho_improvement:.3f}"
            f"({old_rho_max:.2f} -> {min_rho:.2f}) [time: {time.time() - start_time:.02f}s]"
        )
    else:
        logging.info(
            f"t={observation.get_time_stamp()}, line={observation.rho.argmax():03d} overflowed" f"=> No action found!"
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)



"""This file consist of the N-1 Tutor, that executes the actions similar to the N-1 Teacher.

"""
import logging
import os
import random
import time
from datetime import datetime
from multiprocessing import Pool
from pathlib import Path
from typing import Optional, Union, Tuple, List

import grid2op
import numpy as np
from grid2op.Agent import BaseAgent
from grid2op.Environment import BaseEnv
from grid2op.Exceptions import DivergingPowerFlow
from grid2op.dtypes import dt_int

from curriculumagent.common.utilities import find_best_line_to_reconnect
from curriculumagent.tutor.tutors.general_tutor import GeneralTutor


def collect_divergent_powerflow_experience(
        action_paths: Union[Path, List[Path]],
        chronics_id: int,
        env_name_path: Union[Path, str] = "l2rpn_neurips_2020_track1_small",
        seed: Optional[int] = None,
        enable_logging: bool = True,
        runs_per_chronic: int = 100,
        TutorAgent: BaseAgent = GeneralTutor,
) -> List[np.array]:
    """Collect teacher_experience of the tutor right before the DivergingPowerFlow error. For that we run
    through multiple iterations of the environment, in order to collect as much information
    within previously observation to detect the DivergingPowerFlow in advance.

    In this run, we save both the T-1 and T-3 observation, with T being the moment when the agent
    hits a diverging powerflow.

    Args:
        action_paths: List of Paths for the tutor.
        chronics_id: Number of chronic to run.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        seed: Whether to init the numpy seed which is used for the Grid2Op seeds.
        runs_per_chronic: How many times the chronic should be iterated with different seeds.
        enable_logging: Whether to log the Tutor teacher_experience search.
        TutorAgent: Tutor Agent which should be used for the search.

    Returns:
        Returns two array containing the T-1 and T-3 observations of all cases where a Diverging Powerflow
        occurred.
    """
    if enable_logging:
        logging.basicConfig(level=logging.INFO)

    try:
        # if lightsim2grid is available, use it.
        from lightsim2grid import LightSimBackend

        backend = LightSimBackend()
        env = grid2op.make(dataset=env_name_path, backend=backend)
    except ImportError:  # noqa
        env = grid2op.make(dataset=env_name_path)
        logging.warning("Not using lightsim2grid! Operation will be slow!")

    if seed:
        np.random.seed(seed)

    # After initializing the environment, let's init the tutor
    tutor = TutorAgent(action_space=env.action_space, action_space_file=action_paths)

    max_int = np.iinfo(dt_int).max
    env_seeds = list(np.random.randint(max_int, size=runs_per_chronic))
    print(env_seeds)

    logging.info(f"Run current chronic:{env.chronics_handler.get_name()} " f"with a total of {runs_per_chronic} seeds.")

    # We run through multiple itartions of the chronic:
    records_t_minus_1 = []
    records_t_minus_3 = []
    for env_seed in env_seeds:
        env.set_id(chronics_id)
        env.seed(env_seed)
        env.reset()

        done, obs, info = False, env.get_obs(), []
        obs_lists = []
        act_list = []
        while not done:
            action, idx = tutor.act_with_id(obs)

            act = find_best_line_to_reconnect(obs=obs, original_action=env.action_space.from_vect(action))
            obs, _, done, info = env.step(act)

            if not done:
                # Save the last three steps:
                obs_lists.append(obs.copy())
                del obs_lists[:-3]
                act_list.append(idx)
                del act_list[:-3]

        if isinstance(info["exception"], list) and len(info["exception"]) > 0:
            if isinstance(info["exception"][0], DivergingPowerFlow):
                logging.info(f"Divergin Powerflow detected at step {obs.current_step}")
                records_t_minus_1.append(
                    np.hstack([act_list[-1], obs_lists[-1].to_vect()]).astype(np.float32).reshape(1, -1)
                )
                records_t_minus_3.append(
                    np.hstack([act_list[-3], obs_lists[-3].to_vect()]).astype(np.float32).reshape(1, -1)
                )

    return [np.array(records_t_minus_1), np.array(records_t_minus_1)]


def generate_divergent_exp(
        env_name_path: Union[Path, str],
        save_path: Union[Path, str],
        action_paths: Union[Path, List[Path]],
        num_chronics: Optional[int] = None,
        num_sample: Optional[int] = None,
        jobs: int = -1,
        seed: Optional[int] = None,
        TutorAgent: BaseAgent = GeneralTutor,
):
    """Method to run the Divergent Powerflow Search in parallel. Quite similar to the tutor.

    Args:
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        save_path: Where to save the teacher_experience.
        action_paths: List of action sets (in .npy format).
        num_chronics: Total numer of chronics.
        num_sample: Length of sample from the num_chronics. If num_sample is smaller than num chronics,
        a subset is taken. If it is larger, the chronics are sampled with replacement.
        jobs: Number of jobs in parallel.
        seed: Whether to set a seed to the sampling of environments.
        TutorAgent: Tutor Agent which should be used for the search, default is the GeneralTutor.

    Returns:
        None, saves results as numpy file.

    """
    log_format = "(%(asctime)s) [%(name)-10s] %(levelname)8s: %(message)s [%(filename)s:%(lineno)s]"
    logging.basicConfig(level=logging.INFO, format=log_format)

    if jobs == -1:
        jobs = os.cpu_count()

    tasks = []

    # Make sure we can initialize the environment
    # This also makes sure that the environment actually exits or gets downloaded
    env: BaseEnv = grid2op.make(env_name_path)
    chronics_path = env.chronics_handler.path
    if chronics_path is None:
        raise ValueError(f"Can't determine chronics path of given environment {env_name_path}")

    if num_chronics is None:
        num_chronics = len(os.listdir(chronics_path))

    if num_sample:
        if num_sample <= num_chronics:
            sampled_chronics = random.sample(range(num_chronics), num_sample)
        else:
            sampled_chronics = random.choices(np.arange(num_chronics), k=num_sample)
    else:
        sampled_chronics = np.arange(num_chronics)

    for chronic_id in sampled_chronics:
        tasks.append((action_paths, chronic_id, env_name_path, seed, TutorAgent))
    if jobs == 1:
        # This makes debugging easier since we don't fork into multiple processes
        logging.info(f"The following {len(tasks)} tasks will executed sequentially: {tasks}")
        out_result = []
        for task in tasks:
            out_result.append(collect_divergent_powerflow_experience(*task))
    else:
        logging.info(f"The following {len(tasks)} tasks will be distributed to a pool of {jobs} workers:")
        start = time.time()
        with Pool(jobs) as p:
            out_result = p.starmap(collect_divergent_powerflow_experience, tasks)
        end = time.time()
        elapsed = end - start
        logging.info(f"Time: {elapsed}s")

    # Now concatenate the result:
    all_experience = np.concatenate(out_result, axis=0)
    if save_path.is_dir():
        now = datetime.now().strftime("%d%m%Y_%H%M%S")
        save_path = save_path / f"divergent_powerflow_{now}.npy"

    np.save(save_path, all_experience)
    logging.info(f"Divergent PF teacher_experience has been saved to {save_path}")



"""In this file, the code for the hyperparameter search of the junior model is provided.
It is optimized to run the NNI framework @https://nni.readthedocs.io/en/stable/
In order to run this code via NNI, you have to start it via the terminal.
NNI then automatically runs through the parameters and selects the best option.
"""
import logging
import os
from pathlib import Path
from typing import Optional

import nni
import numpy as np
from nni import Experiment
from nni.experiment.config import ExperimentConfig
from sklearn.preprocessing import StandardScaler

from curriculumagent.junior.junior_student import Junior
from curriculumagent.junior.junior_student import load_dataset


def get_default():
    """Method to get the range of hyperparameters.

    Args:
        None.

    Returns:
        Default dictionary of hyperparameters.

    """
    hyper_config = {
        "epochs": 10,  # set to 1000, should be adjusted by the A. hyperband
        "learning_rate": 1e-5,
        "activation": "relu",
        "layer1": 1000,
        "layer2": 1000,
        "layer3": 1000,
        "layer4": 1000,
        "dropout1": 0.25,
        "dropout2": 0.25,
        "patience": 0,
        "initializer": "O",
        'batchsize': 256
    }
    return hyper_config


def run_one_experiment(config,action_space_file, path_to_files: Path, run_nni: bool = True):
    """ Method for the hyper-optimization training with tune. This method will be executed
    with the config file. The config is specified by tune.

    Args:
        action_space_file:
        config: Config file with all the necessary parameters for the Junior model.
        run_nni: Boolean indicator, whether to run the code with NNI. Tip: Test this code first without NNI.

    Returns:
        The accuracy of the validation run.

    """
    config["patience"] = 0
    name = [a for a in os.listdir(path_to_files) if "_train.npz" in a][0][:-10]
    logging.info(f"Assuming the name {name} for the data files")
    s_train, a_train, s_validate, a_validate, _, _ = load_dataset(path_to_files, dataset_name=name)

    # Standardize:
    scaler = StandardScaler()
    s_tr_t = scaler.fit_transform(s_train)
    s_val_t = scaler.transform(s_validate)

    # First initialize the advanced junior:
    # Note: We do not set a seed, to have more variability!!
    junior = Junior(action_space_file=action_space_file,config=config, run_nni=run_nni)

    history = junior.train(
        log_dir=None,
        ckpt_dir=None,
        patience=None,
        x_train=s_tr_t,
        y_train=a_train,
        x_validate=s_val_t,
        y_validate=a_validate,
    )
    acc = max(history.history["val_accuracy"])

    if run_nni:
        nni.report_final_result(acc)
    # we return the best validation result:
    return acc


def run_nni_experiment(path_to_files: Path,
                       max_trial_number: Optional[int] = 100,
                       experiment_working_directory: Optional[Path] = None):
    """ This is the method to run the Hyper-parameter search with NNI based on https://nni.readthedocs.io/.
    In the method, we initialize a model and then run the script within a python method.

    Args:
        path_to_files: Directory where to find the dataset of the tutor, which is then loaded with load_dataset.
        path_to_config: Optional directory, where to find the config of the experiment. If not supplied, we use the
            directory of this file.
        experiment_working_directory: Where the experiment files should be saved.

    Note:
          This method is still experimental and depends on the machine and your directory. A working alternative
          is the start of the experiment via nnictl in the terminal

    Returns:
        None
    """
    exp_config = ExperimentConfig(experiment_name='Junior',
                                  experiment_type="hpo",
                                  trial_concurrency=4,
                                  max_experiment_duration='48h',
                                  max_trial_number=max_trial_number,
                                  training_service_platform='local',
                                  search_space_file=Path(__file__).parent / 'search_space.json',
                                  use_annotation=False,
                                  advisor={'name': 'BOHB',
                                           'class_args': {'optimize_mode': 'maximize',
                                                          'min_budget': 1,
                                                          'max_budget': 100,
                                                          'eta': 3,
                                                          'min_points_in_model': 17,
                                                          'top_n_percent': 20,
                                                          'num_samples': 128,
                                                          'random_fraction': 0.33,
                                                          'bandwidth_factor': 3.0,
                                                          'min_bandwidth': 0.001}},
                                  trial_command=f'python hyper_parameter_search_junior_nni.py --path_to_files={path_to_files}',
                                  trial_code_directory=Path(__file__).parent,
                                  trial_gpu_number=0)

    assert any(
        [".npz" in p for p in os.listdir(path_to_files)]), "Did not find any .npz file, i.e., the tutor experience!"

    # Init Experiment:
    if experiment_working_directory:
        assert experiment_working_directory.is_dir(), f"The dir {experiment_working_directory} does not exist!"
        exp_config.experiment_working_directory = experiment_working_directory

    experiment = Experiment(exp_config)
    logging.info(f"Run the NNI experiment with the following config: {experiment.config}")
    experiment.run(8080, wait_completion=True)
    experiment.stop()
    logging.info("Reverting working directory")


if __name__ == "__main__":
    """ This main method is needed to run the NNI experiment. 
    Either you start the experiment with the run_nni_experiment method, or via the terminal.

    NNI then automatically runs throught the parameters and selects the best option.
    See https://nni.readthedocs.io/en/stable/ for the information

    If you want to test the code above set run_nni=False
    """
    import argparse

    parser = argparse.ArgumentParser(description='Run NNI Experiment')
    parser.add_argument('--path_to_files', metavar='path', required=True,
                        help='Path, where to find the junior train.npz files. This is required.')
    args = parser.parse_args()

    run_nni = True
    params = get_default()
    try:
        # get parameters form tuner
        if run_nni:
            tuner_params = nni.get_next_parameter()
            logging.info(tuner_params)
            tuner_params['epochs'] = tuner_params['TRIAL_BUDGET'] * 10 + 5
            params.update(tuner_params)
            logging.info(params)

        run_one_experiment(config=params,
                           path_to_files=args.path_to_files,
                           run_nni=run_nni)

    except Exception as exception:
        logging.exception(exception)
        raise



import grid2op
from grid2op.Action import BaseAction
from grid2op.Environment import Environment
from grid2op.Reward import BaseReward
from grid2op.dtypes import dt_float


class TopoRew(BaseReward):
    """This class executes a specific topological reward. The idea is that the agent receives a
    positive reward for each step and a "strong" positive reward, if the grid is in its original state.
    """

    def __init__(self, min_rew: float = 0.0, max_rew: float = 1.0, proportion: float = 0.5, logger=None):
        """The reward is calculated as following:
        If something illegal happened or game over is achieved, the agent receives the min_rew.
        If the agent survives the step, it receives the max_rew*proportion reward.
        If the agent survives the step and is also in the original topology, the agent receives max_rew.

        Args:
            min_rew: Minimal reward for the agent.
            max_rew: Maximal reward for the agent.
            proportion: Proportion between the reward and the topo reward. We multiply with the proportion
            logger: Logger.

        Returns:
            None.
        """
        BaseReward.__init__(self, logger=logger)
        self.reward_min = dt_float(min_rew)
        self.reward_max = dt_float(max_rew)
        self.proportion = proportion

    def __call__(self, action: grid2op.Action.BaseAction,
                 env: grid2op.Environment.Environment,
                 has_error,
                 is_done,
                 is_illegal,
                 is_ambiguous) -> float:
        """Here the reward is called. Note that we differentiate between the three different
        is_done options. If is_done=False, then the agent receives the reward. If the agent completed
        the whole env, it also receives the max_rew. If the agent does not complete the env, thus
        is_done = True, it receives the min_rew.

        Args:
            action: Action of the agent.
            env: Grid2Op Environment.
            has_error: Does the Environment Return an Error.
            is_done: Is the Environment Done.
            is_illegal: Is the action illegal.
            is_ambiguous: Is the action of the agent ambiguous.

        Returns:
            Float.

        """
        # check for errors
        if has_error or is_illegal or is_ambiguous:
            return self.reward_min

        # Check dones:
        if is_done:
            if env.nb_time_step == env.max_episode_duration():
                return self.reward_max
            else:
                return self.reward_min

        # Check topology:
        if all(env.current_obs.topo_vect==1):
            return self.reward_max
        else:
            return self.reward_max*self.proportion


class PPO_Reward(BaseReward):
    def __init__(self):
        """
        PPO_Reward class, based on the BaseReward from Grid2Op
        """
        BaseReward.__init__(self)
        self.reward_min = -10
        self.reward_std = 2

    def __call__(
            self, action: BaseAction, env: Environment, has_error: bool, is_done: bool, is_illegal: bool,
            is_ambiguous: bool
    ) -> float:
        """Call of the Reward class. Here the BaseReward is overwritten.

        Compute the reward based on:

        If the action is valid, the reward is calculated as followed:
                2 - rho_max * (1 if rho_max < 0.95 else 2)

        Args:
            action: action of the environment in Grid2Op format
            env: Grid2Op Environment
            has_error: Has there been an error, for example a :class:`grid2op.DivergingPowerFlow` be thrown when the
                action has been implemented in the environment.
            is_done: Is the episode over (either because the agent has reached the end, or because
                there has been a game over)
            is_illegal: Has the action submitted by the BaseAgent raised an
             :class:`grid2op.Exceptions.IllegalAction` exception. In this case it has been replaced by "do nohting" by
             the environment. **NB** an illegal action is NOT an ambiguous action. See the description of the Action
             module: :ref:`Illegal-vs-Ambiguous` for more details. is_ambiguous:  Has the action submitted by the
             BaseAgent raised an :class:`grid2op.Exceptions.AmbiguousAction` exception. In this case it has been
              replaced by "do nothing" by the environment. **NB** an illegal action is NOT an ambiguous action.
              See the description of the Action module: :ref:`Illegal-vs-Ambiguous` for more details.

        Returns: Return, based on the formular

        """
        if is_done or is_illegal or is_ambiguous or has_error:
            return self.reward_min
        rho_max = env.get_obs().rho.max()
        return self.reward_std - rho_max * (1 if rho_max < 0.95 else 2)


"""This file constist of two custom models, which transfer the junior weights into the Rllib experiment.

"""
import warnings
from pathlib import Path
from typing import List, Tuple

import numpy as np
import tensorflow as tf
from gymnasium.spaces import Discrete, Box
from ray.rllib.models.tf.tf_modelv2 import TFModelV2

class Grid2OpCustomModel(TFModelV2):
    """Custom model for policy gradient algorithms.

    """

    def __init__(
        self,
        obs_space: Box,
        action_space: Discrete,
        num_outputs: int,
        model_config: dict,
        name: str,
        **customized_model_kwargs
    ):
        """Constructor of the custom model.

        This is preferably a junior model, however,  a pretrained Senior model should work as well.
        NOte that when using a Senior model, the second option becomes obsolete.

        Args:
            obs_space: Observation space passed by rllib.
            action_space: Action space passed by rllib.
            num_outputs: Number of output passed by rllib, shape of action space.
            model_config: Configurations of the model for RLlib to init the model.
            name: Name of the model, if wanted.
            customized_model_kwargs: Custom Model config from junior Hyper-parameter selections. Here you plug in
            any layer information, activation method or initializer.

        Returns:
            None.

        """
        super(Grid2OpCustomModel, self).__init__(obs_space, action_space, num_outputs, model_config, name)
        if "custom_config" in customized_model_kwargs.keys():
            cconfig = customized_model_kwargs["custom_config"]
        else:
            cconfig = {}
        layer_size, initializer, activation = self._extract_config(cconfig)

        # Now Init model:
        self.inputs = tf.keras.layers.Input(shape=obs_space.shape, name="observations")

        layer1 = tf.keras.layers.Dense(
            layer_size[0], name="layer_1", activation=activation, kernel_initializer=initializer)(
            self.inputs
        )
        layer2 = tf.keras.layers.Dense(layer_size[1], name="layer_2", activation=activation, kernel_initializer=initializer)(
            layer1
        )
        layer3 = tf.keras.layers.Dense(
            layer_size[2], name="layer_3", activation=activation, kernel_initializer=initializer
        )(layer2)
        layer4 = tf.keras.layers.Dense(
            layer_size[3], name="layer_4", activation=activation, kernel_initializer=initializer
        )(layer3)
        act_layer = tf.keras.layers.Dense(
            num_outputs, name="action_out", activation=None, kernel_initializer=initializer
        )(layer4)
        val_hidden_layer = tf.keras.layers.Dense(
            action_space.n, name="layer_val_hidden", activation=activation, kernel_initializer=initializer
        )(layer4)
        val_layer = tf.keras.layers.Dense(1, name="value_out", activation=None, kernel_initializer=initializer)(
            val_hidden_layer
        )

        self.base_model = tf.keras.Model(self.inputs, [act_layer, val_layer])

        path_to_junior = customized_model_kwargs["model_path"]

        self._params_copy(path=path_to_junior)

    def _extract_config(self, config) -> Tuple[List, tf.keras.initializers.Initializer, tf.keras.layers.Layer]:
        """
        Method to extract the hyperparameters of the config

        Args:
            config:

        Returns: layers, initializer and activation methode.

        """
        # Default values:
        layer_size = [1000, 1000, 1000, 1000]
        initializer = tf.keras.initializers.Orthogonal()
        activation = tf.keras.layers.ReLU()

        if not any([k in config.keys() for k in ["activation", "initializer", "layer1",
                                                 "layer2", "layer3", "layer4"]]):
            import warnings
            warnings.warn("The custom dictionary had not the correct keys. Using default model.")
            return layer_size, initializer, activation

        # Activation options:
        activation_option = {"leaky_relu": tf.keras.layers.LeakyReLU(),
                             "relu": tf.keras.layers.ReLU(),
                             "elu": tf.keras.layers.ELU(),
                             }
        # Initializer Options
        initializer_option = {"O": tf.keras.initializers.Orthogonal(),
                              "RN": tf.keras.initializers.RandomNormal(),
                              "RU": tf.keras.initializers.RandomUniform(),
                              "Z": tf.keras.initializers.Zeros()}

        if "activation" in config.keys():
            activation = activation_option[config["activation"]]

        if "initializer" in config.keys():
            initializer = initializer_option[config["initializer"]]

        if all([l in config.keys() for l in ["layer1", "layer2", "layer3", "layer4"]]):
            layer_size = [int(np.round(config[l])) for l in ["layer1", "layer2", "layer3", "layer4"]]

        return layer_size, initializer, activation

    def _params_copy(self, path: Path):
        """Private Method from PPO code. Overwriting the weights of the rllib model by the Junior model.

        This method does one of two things:
        1. Used to copy the weights of the Junior model onto the model of the Rllib custom model.
        2. If the model is already a PPO model, it just copies it.

        Note that for the second option, the custom config does not have an effect.

        Args:
            path: Path of the Junior model checkpoints.

        Returns:
            None.

        """
        model = tf.keras.models.load_model(path)

        if isinstance(model,tf.keras.models.Sequential):
            self.base_model.layers[1].set_weights(model.layers[0].get_weights())
            self.base_model.layers[2].set_weights(model.layers[1].get_weights())
            self.base_model.layers[3].set_weights(model.layers[2].get_weights())
            self.base_model.layers[4].set_weights(model.layers[4].get_weights())
            self.base_model.layers[5].set_weights((*map(lambda x: x / 5, model.layers[6].get_weights()),))
        else:
            model.compile()
            self.base_model = model

    def forward(self, input_dict, state, seq_lens):
        model_out, self._value_out = self.base_model(input_dict["obs"])
        return model_out, state

    def value_function(self):
        return tf.reshape(self._value_out, [-1])

    def metrics(self):
        return {"foo": tf.constant(42.0)}



"""This file is build to construct the environment of the Senior Agent to run the PPO with ray and rllib.
"""
import logging
import os
import pickle
from pathlib import Path
from typing import Union, TypedDict, Tuple, List, Optional, Dict

import grid2op
import gymnasium as gym
import numpy as np
import ray
from grid2op.Exceptions import UnknownEnv
from gymnasium.spaces import Discrete, Box
from lightsim2grid import LightSimBackend
from ray._raylet import ObjectRef
from sklearn.base import BaseEstimator

from curriculumagent.common.obs_converter import obs_to_vect
from curriculumagent.common.utilities import find_best_line_to_reconnect, split_and_execute_action, revert_topo
from curriculumagent.senior.rllib_execution.alternative_rewards import PPO_Reward


class RLlibEnvConfig(TypedDict):
    """TypeDict Class.

     Attributes:
        action_space_path: Either path to the actions or a list containing mutliple actions.
        env_path: Path of the Grid2Op environment
        action_threshold: Between 0 and 1.
        subset: Should the obs.to_vect be filtered similar to the original Agent. If True,
                the obs.to_vect is filtered based on predefined values. If False, all values are considered.
                Alternatively, one can submit the once own values.
                testing: Indicator, whether the underlying Grid2op Env should be started in testing mode or not
        scaler: Optional Scaler of Sklearn Model. Either a Sklearn Scaler or its ray ID, if the scaler is saved via
                ray.put(). If scaler is provided, the environment will scale the observations based on
                scaler.transform().
        env_kwargs: Optional additional arguments for the Grid2Op environment that should be used when making the
        environment.

    """
    action_space_path: Union[Path, List[Path]]
    env_path: Union[str, Path]
    action_threshold: float
    filtered_obs: Union[bool, List]
    scaler: Optional[Union[ObjectRef, BaseEstimator, str]]
    topo: Optional[bool]
    alternative_rew: Optional[grid2op.Reward.BaseReward]
    env_kwargs: Optional[dict] = {}


class SeniorEnvRllib(gym.Env):
    """Environment class, in which the Grid2Op Backend is combined with the actions of the tutor agent.
    The environment can then be used in the RLlib framework for training.

    There is a difference between this environment and the GymEnv of grid2op.gym_compat, because we
    still access the Grid2Op Environment.

    Note: With the Update to Ray>=2.4 the environment is now based on the gymnasium package and
        not gym anymore. Thus, small changes are necessary.


    """

    def __init__(self, config: RLlibEnvConfig,
                 testing: bool = False):
        """Initialization of Environment with Grid2Op.

        If possible, the lightsim2grid backend is used. Further, we define based on the Gym format
        the environment.

        Args:
            config: RllibEnvConfig File, containing the following keys,value pairs:
                action_space_path: Either path to the actions or a list containing mutliple actions.
                env_path: Path of the Grid2Op environment
                action_threshold: Threshold between 0 and 1.
                subset: Should the obs.to_vect be filtered similar to the original Agent. If True,
                        the obs.to_vect is filtered based on predefined values by the obs_to_vect method.
                        If False, all values are considered. Alternatively, one can submit the own values.
                        testing: Indicator, whether the underlying Grid2op Env should be started in testing mode or not
                scaler: Optional Scaler of Sklearn Model. Either a Sklearn Scaler or its ray ID, if the scaler is saved
                        ray.put(). If scaler is provided, the environment will scale the observations based on
                        scaler.transform().
                alternative_rew: If wanted, you can supply an alternative reward for the grid2op Environment.

        Returns:
            None.

        """
        # Check if subsample of Obs:
        action_space_path = config["action_space_path"]
        action_threshold = config["action_threshold"]

        self.subset = config["subset"]

        data_path = str(config["env_path"])
        env_kwargs = config.get("env_kwargs", {})

        # Init environments: If testing is true, we test the test flag
        backend = LightSimBackend()
        try:
            if testing:
                env = grid2op.make(dataset=data_path, backend=backend,
                                   reward_class=PPO_Reward, test=True, **env_kwargs)
                logging.info("Test flag was set to true. Init Env in testing mode.")
            else:
                if "alternative_rew" in config.keys():
                    env = grid2op.make(dataset=data_path, backend=backend,
                                       reward_class=config["alternative_rew"], test=False, **env_kwargs)
                else:
                    env = grid2op.make(dataset=data_path, backend=backend,
                                       reward_class=PPO_Reward, test=False, **env_kwargs)
                logging.info("Init of Grid2Op Env works.")
            self.single_env = env.copy()
        except UnknownEnv as test_env:
            logging.info(f"Error testing was raised {test_env}. Assume testing, thus set test flag.")
            self.single_env = grid2op.make(dataset=data_path, test=True,
                                           backend=LightSimBackend(), reward_class=PPO_Reward)

        self.single_env.chronics_handler.shuffle(
            shuffler=lambda x: x[np.random.choice(len(x), size=len(x), replace=False)])
        # Create multiple environments:
        self.obs_grid2op = self.single_env.reset()

        # Initialize the actions:
        if isinstance(action_space_path, Path):
            if action_space_path.is_file():
                logging.info(f"Action_space_path {action_space_path} is a file and will be loaded.")
                self.actions = np.load(str(action_space_path))
            elif action_space_path.is_dir():
                logging.info(f"Action_space_path {action_space_path} is a path. All available action files "
                             f" will be loaded.")
                all_action_files = [act for act in os.listdir(action_space_path) if "actions" in act and ".npy" in act]

                if not all_action_files:
                    raise FileNotFoundError("No actions files were found!")

                loaded_files = [np.load(str(action_space_path / act)) for act in all_action_files]
                self.actions = np.concatenate(loaded_files, axis=0)

        elif isinstance(action_space_path, list):
            logging.info(f"Action_space_path {action_space_path} is a list containing multiple actions.")
            loaded_files = [np.load(str(act_path)) for act_path in action_space_path]
            self.actions = np.concatenate(loaded_files, axis=0)
        else:
            raise ValueError(f"The action_space_path variable {action_space_path} does neither consist of a single "
                             f"action nor of a path where actions can be found.")

        # Define Action and Observation Space in Gym Format:
        self.action_space = Discrete(len(self.actions))

        if isinstance(self.subset, list):
            vect_shape = len(self.subset)
        elif self.subset:
            vect_shape = obs_to_vect(self.single_env.get_obs(), False).shape[0]
        else:
            vect_shape = self.single_env.observation_space.size()
        self.observation_space = Box(shape=(vect_shape,), high=np.inf, low=-np.inf)

        logging.info(f"Initialize Environment with the following action space{self.action_space}  and the following "
                     f"observations space{self.observation_space}")

        # check scaler:
        self.scaler = config.get("scaler", None)
        if self.scaler is not None:
            if isinstance(self.scaler, (str, Path)):
                with open(self.scaler, "rb") as fp:
                    self.scaler = pickle.load(fp)

        self.topo = config.get("topo", False)

        # Initialize further arguments:
        assert 0.0 < action_threshold <= 1.0
        self.action_threshold = action_threshold

        self.step_in_env = 0
        self.passiv_reward = 0
        self.obs_rl = self.run_next_action()
        self.reward = None

        self.info: Dict = {}
        self.done = self.single_env.done
        # Added for gymnasium
        self.truncated, self.terminated = self.__check_terminated_truncated()

    def run_next_action(self) -> np.ndarray:
        """Run the environment until an action by the agent is required.

        In this function, the Grid2Op Environment will run until it requires help
        from the agent. Then it will return the observation and wait for action.

        Returns:
            Observation vector.

        """
        # init values
        continue_running = True
        cur_obs = self.single_env.get_obs()
        cumulated_rew = 0
        # Run through the env and do nothing.
        # The reward is collected until the agent is done.
        # If done, there is an extra reward.
        while continue_running:
            if cur_obs.rho.max() < self.action_threshold:
                if self.topo:
                    action_array = revert_topo(self.single_env.action_space, cur_obs)
                    default_action = self.single_env.action_space.from_vect(action_array)
                else:
                    default_action = self.single_env.action_space({})

                action = find_best_line_to_reconnect(obs=cur_obs,
                                                     original_action=default_action)
                # Execute in env:
                cur_obs, reward, _, self.info = self.single_env.step(action)
                cumulated_rew += reward
                self.step_in_env += 1

                if self.single_env.done:
                    if 'GAME OVER' in str(self.info['exception']):
                        cumulated_rew += -300
                    else:
                        cumulated_rew += 500
                    continue_running = False

            else:
                continue_running = False
        self.obs_grid2op = cur_obs
        self.passiv_reward += cumulated_rew

        # Select subset if wanted
        if isinstance(self.subset, list):
            obs_rl = self.obs_grid2op.to_vect()[self.subset]
        elif self.subset:
            obs_rl = obs_to_vect(self.obs_grid2op, False)
        else:
            obs_rl = self.obs_grid2op.to_vect()

        # Select scaler if wanted
        if self.scaler:
            if isinstance(self.scaler, BaseEstimator):
                obs_rl = self.scaler.transform(obs_rl.reshape(1, -1)).reshape(-1, )
            elif isinstance(self.scaler, ObjectRef):
                obs_rl = ray.get(self.scaler).transform(obs_rl.reshape(1, -1)).reshape(-1, )

        return obs_rl

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, Union[float, int], bool, bool, dict]:
        """Conduct the action of the agent.

        This method does not only run the action of the agent but additionally accumulates the
        reward for the do-nothing agent AFTER the action of the agent. Thus, if the action "saves" a
        lot of steps, the agent receives a handsome reward.

        Note: This method checks, whether the action is a tuple or a triple action. If the action
        requires multiple steps indeed, then the action will be executed sequentially, i.e., 2 or 3 steps
        in the Grid2Op Environment.

        With the update to gymnasium the output is a little bit different

        Args:
            action: ID of the action within the action set.

        Returns:
            The observation as np.ndarray, the reward, the info if it is terminated, the boolean if truncated
            and the info, which is default.

        """
        # For some reason sometime the environment is done but ray does not reinit it
        if self.single_env.done:
            logging.info("The Grid2Op Environment seems to be not initialized. We just return the last"
                         "observation and do not run anything")
            self.done = True
            self.truncated, self.terminated = self.__check_terminated_truncated()

            return self.obs_rl, 0.0, self.terminated, self.truncated, self.info

        cur_obs, reward, self.done, self.info = split_and_execute_action(env=self.single_env,
                                                                         action_vect=self.actions[action])

        # get action:
        self.step_in_env += 1
        self.obs_grid2op = cur_obs

        # Select subset if wanted
        if isinstance(self.subset, list):
            obs_rl = self.obs_grid2op.to_vect()[self.subset]
        elif self.subset:
            obs_rl = obs_to_vect(self.obs_grid2op, False)
        else:
            obs_rl = self.obs_grid2op.to_vect()

        if self.scaler:
            if isinstance(self.scaler, BaseEstimator):
                obs_rl = self.scaler.transform(obs_rl.reshape(1, -1)).reshape(-1, )
            elif isinstance(self.scaler, ObjectRef):
                obs_rl = ray.get(self.scaler).transform(obs_rl.reshape(1, -1)).reshape(-1, )

        self.obs_rl = obs_rl

        if (cur_obs.rho.max() < self.action_threshold) and self.done is False:
            self.obs_rl = self.run_next_action()
            reward += self.passiv_reward

        self.done = self.single_env.done

        # We add the passive reward to the overall reward
        self.reward = reward

        # Now check for Truncation
        self.truncated, self.terminated = self.__check_terminated_truncated()

        return self.obs_rl, self.reward, self.terminated, self.truncated, self.info

    def reset(self, seed: Optional[int] = None, options: Optional[Dict[str, int]] = None) -> Tuple[np.ndarray, dict]:
        """ Resetting the environment.

        Args:
            seed: Optional seed that gets passed to the environment.
            options: Optional id for the Grid2Op Environment. When you want to reset to a specific chronic
            reset the env with a dictionary and the key "id". As example {"id": 42}.

        Returns:
            Tuple with the observation and the info statement

        """
        try:
            if seed:
                self.single_env.seed(seed)
            if isinstance(options, dict):
                if "id" in options.keys():
                    self.single_env.set_id(options)
        except Exception as e:
            logging.warning(f"The ID did not work with the environment and raised {e}. Reset without setting ID")

        # Reset environment and run the Grid2Op Env
        self.done = True
        while self.done:
            # We have to iterate over the environment, because it can happen that the environment
            # can be finished without the intervention of the agent. By this loop we ensure that the
            # Senior is needed.
            self.obs_grid2op = self.single_env.reset()
            self.step_in_env = 0
            self.passiv_reward = 0
            self.obs_rl = self.run_next_action()
            if not self.single_env.done:
                self.done = False

        self.truncated, self.terminated = self.__check_terminated_truncated()

        return self.obs_rl, self.info

    def render(self, mode='human') -> None:
        """Required for GYM Environment. However, we do not change anything.

        Args:
            mode: String.

        Returns:
             None.

        """

    def __check_terminated_truncated(self) -> Tuple[bool, bool]:
        """ Check the env for terminated and truncated

        Returns:

        """

        if self.done:
            trunc, term = True, False
            # Check for termination (and revert truncation)
            if self.single_env.nb_time_step == self.single_env.chronics_handler.max_episode_duration():
                trunc, term = False, True
        else:
            trunc, term = False, False

        return trunc, term



"""
This file runs the tuple and triple search defined in :mod:`topology_action_search` and saves the collected
teacher_experience in a csv file for further processing using :mod:`generate_action_space`.
"""
import logging
import os
import time
from multiprocessing import Pool
from pathlib import Path
from typing import Optional, Union

import defopt
import grid2op
import numpy as np
from grid2op.Action import BaseAction
from grid2op.Environment import BaseEnv
from tqdm import tqdm

from curriculumagent.teacher.submodule.common import save_sample_new
from curriculumagent.common.utilities import find_best_line_to_reconnect
from curriculumagent.teacher.submodule.encoded_action import EncodedTopologyAction
from curriculumagent.teacher.submodule.topology_action_search import topology_search_sequential_x_steps


def collect_sequential_experience(
        save_path: Path,
        env_name_path: Union[str, Path],
        chronics_path: Optional[str],
        steps: int,
        seed: int,
        rho_acting_threshold: float = 0.9,
        actions_dict: Optional[dict] = None,
        chronic_limit: Optional[int] = None,
) -> None:
    """Teacher class which collects the sequential teacher_experience in order to have the best values.

    Args:
        save_path: The path to save the teacher_experience to.
        env_name_path: The name or path of the environment used.
        chronics_path: The path to the chronics of the environment.
        steps: Length of combined actions, e.g. if steps is 3, then sequentially 3 substations are effected.
        seed: The seed of the environment.
        actions_dict: Dictionary containing all available actions with the substations as keys.
        rho_acting_threshold: When the environments max_rho goes above this threshold the agent begins to act.
        chronic_limit: How many chronics to run. Runs endless if set to None.

    Returns:
        None.
    """
    if isinstance(env_name_path, Path):
        env_name_path = str(env_name_path)

    try:
        # if lightsim2grid is available, use it.
        from lightsim2grid import LightSimBackend

        backend = LightSimBackend()
        if chronics_path:
            # Use env_name_path as path together with chronics_path
            env: BaseEnv = grid2op.make(dataset=env_name_path, chronics_path=chronics_path, backend=backend)
        else:
            # Use env_name_path as name
            env: BaseEnv = grid2op.make(dataset=env_name_path, backend=backend)
    except Exception as e:  # noqa
        logging.warning("Not using lightsim2grid! Operation will be slow!")
        raise e

    # Loading the different values
    logging.info("Separate all available action on the respective substations for faster execution.")
    all_available_actions = env.action_space.get_all_unitary_topologies_set(env.action_space)

    if not isinstance(actions_dict, dict):
        actions_dict = {}
        for act in tqdm(all_available_actions):
            sub_effected = act.as_dict()["set_bus_vect"]["modif_subs_id"][0]

            if sub_effected in actions_dict:
                actions_dict[sub_effected].append(act)
            else:
                actions_dict[sub_effected] = [act]
        logging.info(
            f"Done with separating the actions. There is a total of {len(actions_dict.keys())} substations with "
            f"overall {len(all_available_actions)} number of actions."
        )

    # Setup chronics with deterministic shuffling
    np.random.seed(seed)
    env.chronics_handler.shuffle(shuffler=lambda x: x[np.random.choice(len(x), size=len(x), replace=False)])
    num_chronics = len(os.listdir(chronics_path))
    if chronic_limit:
        num_chronics = min(num_chronics, chronic_limit)

    # Create jobs for multiprocessing:
    for _ in range(num_chronics):
        env.reset()  # env.reset() loads the next chronic
        dst_step = 0
        logging.info(f"Scenario to test is [{env.chronics_handler.get_name()}], start from step-{dst_step:d}... ...")
        env.fast_forward_chronics(dst_step)
        obs, done = env.get_obs(), False
        while not done:
            if obs.rho.max() >= rho_acting_threshold:

                # Now run through the sequential topological search. Note that this will continue the
                # environment by x-steps.
                # Note that we do not reconnect the lines to not divert the
                rho_old = obs.rho.max()
                best_actions = topology_search_sequential_x_steps(
                    env=env, sub_action_set=actions_dict, steps=steps, show_progress=False
                )

                done = env.done

                if not done or (done and obs.current_step == obs.max_step):
                    rho_new = env.get_obs().rho.max()
                    rho_improvement: float = rho_old - rho_new
                    if "set_bus_vect" in best_actions.as_dict().keys():
                        save_sample_new(
                            Path(save_path),
                            [(rho_improvement, EncodedTopologyAction(best_actions))],
                            obs,
                            env.get_obs(),
                            top_k=1,
                        )

            else:
                action = env.action_space({})
                action = find_best_line_to_reconnect(obs, action)
                obs, _, done, _ = env.step(action)
            dst_step += 1

        if done and obs.current_step < obs.max_step:
            logging.info(
                f"Game over at t={obs.current_step}/{obs.max_step} at chronic: {env.chronics_handler.get_name()}"
            )


def run_sequential_teacher(
        save_path: Path,
        env_name_path: str = "l2rpn_neurips_2020_track1_small",
        steps: int = 3,
        rho_acting_threshold: float = 0.925,
        *,
        n_episodes: int = 100,
        limit_chronics: int = None,
        jobs: int = os.cpu_count(),
        seed: int = 42,
) -> Path:
    """This teacher is trying to collect useful actions by doing topology search while running an
    environment when an overflow occurs. It also tries to combine actions to form tuple or triple actions.

    Args:
        save_path: Path to save the teacher_experience file to.
        steps: Number of combined sequential steps for the teacher to search for. Default is three sequential steps.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        rho_acting_threshold: When the environments max_rho gues above this threshold the agent begins to act.
        n_episodes: Number of episodes(run through all scenarios) to run.
        limit_chronics: Limit the number of chronics to run through.
        jobs: Number of jobs to use for parallel teacher_experience collection.
        seed: The random seed to use.

    Returns: Saving path.
    """

    env_path = Path(env_name_path)
    chronics_path = None
    if env_path.is_dir():
        env_name_path = env_path
        chronics_path = str(env_name_path / "chronics")

    start = time.time()
    tasks = []
    for episode in range(n_episodes):
        tasks.append(
            (save_path, env_name_path, chronics_path, steps, seed + episode, rho_acting_threshold, None, limit_chronics)
        )

    logging.info(f"The following {len(tasks)} tasks will be distributed to a pool of {jobs} workers: {tasks}")
    # Sequential implementation, useful for debugging more easily
    if jobs == 1:
        for task in tasks:
            collect_sequential_experience(*task)
    else:
        # Parallel implementation
        with Pool(jobs, maxtasksperchild=1) as p:
            p.starmap(collect_sequential_experience, tasks)
    end = time.time()
    elapsed = end - start
    logging.info(f"Needed time: {elapsed}s")

    return save_path




"""This is the advanced way to train the Senior model of the CurriculumAgent pipeline.
The Senior can trains the previous agent or Junior model an is based on the original work of Binbinchen.
Here the model uses RLlib to tune the hyper-parameter via Population Based Training.
"""
import os
from pathlib import Path
import random

import ray
from ray import tune
from ray.rllib.models import ModelCatalog
from ray.tune.schedulers import PopulationBasedTraining
from curriculumagent.senior.rllib_execution.senior_env_rllib import SeniorEnvRllib
from curriculumagent.senior.rllib_execution.senior_model_rllib import Grid2OpCustomModel, AdvancedCustomModel


def train_senior():
    """Wrapper method to train the senior with Ray. Please adjust accordingly.

    Args:
        None.

    Returns:
        None, Saves ray checkpoints.

    """

    # Init all variables and Ray
    NUM_CPUS = os.cpu_count()
    NUM_GPUS = 0  # Change if applicable
    NUM_TRIALS = 3

    DATA_PATH = Path("~/data_grid2op/l2rpn_neurips_2020_track1_small").expanduser()
    ACTION_SPACE_DIRECTORY = "../action_space"
    junior_model = Path(__file__).parent / "JuniorModel_new"

    ray.init(num_cpus=NUM_CPUS, num_gpus=NUM_GPUS)
    env_config = {
        "action_space_path": ACTION_SPACE_DIRECTORY,
        "env_path": DATA_PATH,
        "action_threshold": 0.9,
        "filtered_obs": True,
    }
    model_config = {"path_to_junior": junior_model}
    ModelCatalog.register_custom_model("Senior", Grid2OpCustomModel)

    # Set population based training
    pbt = PopulationBasedTraining(
        time_attr="training_iteration",
        metric="episode_reward_mean",
        mode="max",
        perturbation_interval=2,
        resample_probability=0.5,
        hyperparam_mutations={
            "lr": [1e-3, 5e-4, 1e-4, 5e-5, 1e-5],
            "num_sgd_iter": lambda: random.randint(3, 10),
            "vf_loss_coeff": lambda: random.uniform(0.5, 1),
            "clip_param": lambda: random.uniform(0.01, 0.5),
            "gamma": lambda: random.uniform(0.975, 1),
            "entropy_coeff": lambda: 10 ** -random.uniform(2, 5),
        },
    )

    # Run Ray:
    tune.run(
        "PPO",
        checkpoint_freq=1,
        scheduler=pbt,
        keep_checkpoints_num=5,
        verbose=1,
        max_failures=3,
        num_samples=NUM_TRIALS,  # Adjust number of samples accordingly
        local_dir="~/bm",
        stop={"training_iteration": 10},
        config={
            "env": SeniorEnvRllib,
            "env_config": env_config,
            "num_workers": (NUM_CPUS - NUM_TRIALS) / NUM_TRIALS,
            "num_envs_per_worker": 5,
            "lr": 5e-5,
            "num_gpus": NUM_GPUS / NUM_TRIALS,
            "num_cpus_per_worker": 1,
            "remote_worker_envs": False,
            # "model": {"use_lstm": False, "fcnet_hiddens": [1000, 1000, 1000, 1000], "fcnet_activation": 'relu',
            #           "vf_share_layers": True},
            "model": {"custom_model": "binbinchen", "custom_model_config": model_config},
        },
    )




"""This file constist of common methods used by multiple teacher agents.

Most methods are based on the original code of EI Innovation Lab, Huawei Cloud, Huawei Technologies.

"""

from pathlib import Path
from typing import List, Tuple

import numpy as np
import pandas as pd
from grid2op.Action import BaseAction
from grid2op.Observation import BaseObservation

from curriculumagent.teacher.submodule.encoded_action import EncodedTopologyAction


def save_sample_new(
        save_path: Path,
        best_actions: List[Tuple[float, EncodedTopologyAction]],
        obs: BaseObservation,
        obs_: BaseObservation,
        top_k: int,
):
    """Save the given action taken in the given environment for further investigation and action space reduction.

    Args:
        save_path: The path to the csv file where the teacher_experience should be saved/appended.
        best_actions: The best actions that could be taken.
        obs: Observation before the action.
        obs_: Observation after taking the action.
        top_k: How many of the best_actions should be saved.

    Returns:
        None.
    """
    rho_max_old = obs.rho.max()
    rho_max_new = obs_.rho.max()
    rho_improvement: float = best_actions[0][0]

    best_actions = best_actions[:top_k]
    best_action = best_actions[0][1]
    best_actions_only: List[EncodedTopologyAction] = [action[1] for action in best_actions]
    n_actions = len(best_actions_only)
    for i in range(top_k - n_actions):
        best_actions_only.append(EncodedTopologyAction(None))
    if len(best_actions_only) != top_k:
        print(f"Didn't match: {len(best_actions_only)} != {top_k}")
        raise AssertionError()

    rho_improvement_only = [action[0] for action in best_actions]

    data = np.concatenate(
        (
            np.array((rho_max_old, rho_max_new, rho_improvement, best_action)),
            np.array(best_actions_only),
            np.array(rho_improvement_only),
        )
    ).reshape([1, -1])

    # Save to a default path if save_path only directory
    if save_path.is_dir():
        save_path = save_path / "teacher_experience.csv"

    if not save_path.exists():
        column_names = ["rho_max_old", "rho_max_new", "rho_improvement", "best_action"]
        for i in range(len(best_actions_only)):
            column_names.append(f"top_{i}_action")
        for i in range(len(best_actions_only)):
            column_names.append(f"top_{i}_ri")
        # Write header when first writing to file
        data_df = pd.DataFrame(data, columns=column_names)
        data_df.to_csv(save_path, index=False, header=True, mode="a")
    else:
        # Else just append to file
        data_df = pd.DataFrame(data)
        data_df.to_csv(save_path, index=False, header=False, mode="a")


station_id_maps = {}


def affected_substations(action: BaseAction) -> List[int]:
    """Optimizing routine to get all the affected substations of an action.

    Args:
        action: The action to list the affected stations from.

    Returns: A tuple of affected station ids.
    """
    global station_id_maps
    if action.env_name not in station_id_maps:
        station_id_maps[action.env_name] = make_station_id_lookup_map(action.sub_info)
    station_id_map = station_id_maps[action.env_name]
    topo_vect = action._set_topo_vect  # noqa
    affected_sub_ids = np.sort(np.unique(station_id_map[topo_vect > 0]))
    return list(affected_sub_ids)


def make_station_id_lookup_map(sub_info: np.ndarray) -> np.ndarray:
    """Create lookup table for matching station objects in _set_topo_vect to substation ids.

    Args:
        sub_info: The sub_info vector usually taken from a Grid2Op action.

    Returns:
        A numpy array that maps from _set_topo_vect index to substation id
    """
    n_lines = np.sum(sub_info)
    lookup_map = np.full(n_lines, -1, dtype=int)
    i = 0
    for sub_id, n_obj in enumerate(sub_info):
        for _ in range(n_obj):
            lookup_map[i] = sub_id
            i += 1
    return lookup_map



"""This file consist of teh EncodedTopologyAction class, which is used to convert the teacher_experience of the teacher
to a more cache friendly way.

"""

import base64
import zlib
from typing import Optional

import numpy as np
from grid2op.Action import BaseAction
from grid2op.Environment import BaseEnv


class EncodedTopologyAction:
    """A topology/set_bus action encoded as a zlib-compressed base64 string to save it more easily in a csv.

    Note:
        This only encodes set_bus actions, all other action types like redispatch are currently ignored!
    """

    def __init__(self, action: Optional[BaseAction]):
        """Create and encode the given action to a new instance of this object.

        Args:
            action: The Grid2Op action to encode. If set to None, the do nothing action will be encoded.

        Return:
            None.
        """
        self.data: str = self.encode_action(action)

    def to_action(self, env: BaseEnv) -> BaseAction:
        """Decode the action back to a Orid2Op action.

        Args:
            env: The environment this action belongs to.

        Returns:
            The Grid2Op action usable by an agent.
        """
        return self.decode_action(self.data, env)

    def __hash__(self):
        return hash(self.data)

    def __str__(self):
        return self.data

    @staticmethod
    def encode_action(action: Optional[BaseAction]) -> str:
        """Pack a set_bus action into a base64 string to make it hashable and more efficient to save in a .csv file.

        Args:
            action: The Grid2Op action to encode. If set to None, the do nothing action will be encoded.

        Returns:
            An utf8 string containing a base64 encoded representation of the change_bus action.
        """
        # Check if the given action can be encoded.
        if not action:
            return "0"
        assert not (
                action._modif_inj
                and action._modif_change_bus
                and action._modif_set_status
                and action._modif_change_status
                and action._modif_redispatch
                and action._modif_storage
                and action._modif_curtailment
                and action._modif_alarm
        ), "Given action type can be encoded"
        # Special case: Empty action -> Encode with 0
        if not action._modif_set_bus:
            return "0"

        packed_action = zlib.compress(action._set_topo_vect, level=1)
        encoded_action = base64.b64encode(packed_action)
        return encoded_action.decode("utf-8")

    @staticmethod
    def decode_action(act_string: str, env: BaseEnv) -> BaseAction:
        """Unpack the previously encoded string to na action for the given environment.

        Args:
            act_string: The string containing the encoded action.
            env: The environment this action belongs to.

        Returns:
            The Grid2Op action usable by an agent.
        """
        unpacked_act: BaseAction = env.action_space()
        # Special case: Empty action
        if act_string == "0":
            return unpacked_act
        decoded = base64.b64decode(act_string.encode("utf-8"))
        unpacked = np.frombuffer(zlib.decompress(decoded), dtype=np.int32)
        unpacked_act._set_topo_vect = unpacked
        unpacked_act._modif_set_bus = True
        return unpacked_act



# Provide version-number etc.
# This file is automatically edited by git!
# This file must not have any sideeffects or do anything but define meta variables!
from ._version import get_versions

__version__ = get_versions()['version']
__copyright__ = '2023, Fraunhofer IEE'
__title__ = 'CurriculumAgent'
__description__ = 'CurriculumAgent is a cleanup and improved version of the NeurIPS 2020 Competition Agent by ' \
                  'Binbinchen.The agent is build to extract action sets of the Grid2Op Environment and then ' \
                  'use rule-based agent to train a Reinforcement Learning agent.'
__author__ = 'OE224'
__authors__ = ['Malte Lehna']



"""In this file, we do the following thing repeatedly:
    1. choose a scenario
    2. sample a time-step every 6 hours
    3. disconnect a line which is under possible attack
    4. search a greedy action to minimize the max rho (~60k possible actions)
    5. save the tuple of (attacked line, observation, action) to a csv file.

Credit: The methods are the enhanced methods of the original code, see
@https://github.com/AsprinChina/L2RPN_NIPS_2020_a_PPO_Solution
"""
import logging
import os
import random
import time
from multiprocessing import Pool
from pathlib import Path
from typing import Union, Optional

import defopt
import grid2op
import numpy as np
from grid2op.Backend import PandaPowerBackend
from grid2op.Environment import BaseEnv

from curriculumagent.teacher.submodule.common import save_sample_new
from curriculumagent.teacher.submodule.encoded_action import EncodedTopologyAction
from curriculumagent.teacher.submodule.topology_action_search import topology_search_topk


def collect_attacker_experience(
        save_path: Path,
        env_name_path: Union[str, Path],
        chronics_path: str,
        line_to_disconnect: int,
        n_episodes: int,
        top_k: int,
        seed: Optional[int] = None,
):
    """Collect useful actions by attacking/disconnecting specific lines at random timesteps
    and trying to find the best action to resolve that problem.

    Args:
        save_path: Path to save the teacher_experience file to.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        line_to_disconnect: The line to disconnect in this trial/scenario.
        chronics_path: The path to the chronics of the environment.
        n_episodes: Number of episodes(run through all scenarios) to run.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        seed: Optional seed for Grid2Op Env and sampling.

    Returns:
        None.
    """
    # Setup environment with the best possible backend
    try:
        # if lightsim2grid is available, use it.
        from lightsim2grid import LightSimBackend

        backend = LightSimBackend()
    except ImportError:  # noqa
        backend = PandaPowerBackend()
        logging.warning("Not using lightsim2grid! Operation will be slow!")
    if chronics_path:
        # Use env_name_path as path together with chronics_path
        env: BaseEnv = grid2op.make(dataset=env_name_path, chronics_path=chronics_path, backend=backend)
    else:
        # Use env_name_path as name
        env: BaseEnv = grid2op.make(dataset=env_name_path, backend=backend)

    if seed:
        np.random.seed(seed)
        env.seed(seed)

    env.chronics_handler.shuffle(shuffler=lambda x: x[np.random.choice(len(x), size=len(x), replace=False)])
    # traverse all scenarios
    all_actions = env.action_space.get_all_unitary_topologies_set(env.action_space)
    for _ in range(len(os.listdir(chronics_path))):
        env.reset()  # env.reset() loads the next chronic
        timestep = n_episodes * 72 + random.randint(0, 72)  # a random sampling every 6 hours
        logging.info(
            f"Scenario[{env.chronics_handler.get_name()}]: "
            f"at step[{timestep:d}], disconnect line-{line_to_disconnect:d}"
            f"(from bus-{env.line_or_to_subid[line_to_disconnect]:d} "
            f"to bus-{env.line_ex_to_subid[line_to_disconnect]:d}]"
        )
        # to the destination time-step
        env.fast_forward_chronics(timestep - 1)
        obs, _, done, _ = env.step(env.action_space({}))
        if done:
            break
        # disconnect the targeted line
        new_line_status_array = np.zeros(obs.rho.shape, dtype=int)
        new_line_status_array[line_to_disconnect] = -1
        action = env.action_space({"set_line_status": new_line_status_array})
        obs, _, done, _ = env.step(action)
        if obs.rho.max() < 1:
            # not necessary to do a dispatch
            continue

        # search a greedy action
        best_actions = topology_search_topk(env, obs, all_actions, top_k=125)
        action = best_actions[0][1]
        best_actions = [(ri, EncodedTopologyAction(act)) for ri, act in best_actions]
        obs_, _, done, _ = env.step(action)
        save_sample_new(Path(save_path), best_actions, obs, obs_, top_k=top_k)


def run_attacking_teacher(
        save_path: Path,
        env_name_path: Union[Path, str] = "l2rpn_neurips_2020_track1_small",
        *,
        n_episodes: int = 100,
        top_k: int = 125,
        jobs: int = os.cpu_count(),
        seed: Optional[int] = None,
):
    """This teacher is trying to collect useful actions by attacking/disconnecting specific lines at random timesteps
    and trying to find the best action to resolve that problem.

    Args:
        save_path: Path to save the teacher_experience file to.
        env_name_path: Path to Grid2Op dataset or the standard name of it.
        n_episodes: Number of episodes(run through all scenarios) to run.
        top_k: How many of the top best actions should be saved in the teacher_experience file at save_path.
        jobs: Number of jobs to use for parallel teacher_experience collection.
        seed: Optional seed for the Grid2Op Environment and random sampling.

    Returns:
        None.
    """
    # hyper-parameters
    env_lines2attack = {
        "l2rpn_case14_sandbox": [1, 10, 15],
        "l2rpn_neurips_2020_track1_small": [45, 56, 0, 9, 13, 14, 18, 23, 27, 39],
    }

    env_name = env_name_path
    # Try to resolve env_name using path, e.g. ~/data_grid2op/env_name -> env_name
    for key in env_lines2attack:
        if key in str(env_name_path):
            env_name = key

    try:
        lines2attack = env_lines2attack[env_name]
    except KeyError as error:
        raise ValueError(f"No lines to attack available for {env_name}") from error

    env_name_path = Path(env_name_path).expanduser()
    chronics_path = None
    if env_name_path.is_dir():
        chronics_path = str(env_name_path / "chronics")
    env_name_path = str(env_name_path)

    start = time.time()
    # Parallel implementation
    tasks = []
    for episode in range(n_episodes):
        for line_to_disconnect in lines2attack:
            tasks.append((save_path, env_name_path, chronics_path, line_to_disconnect, episode, top_k, seed))
    logging.info(f"The following {len(tasks)} tasks will be distributed to a pool of {jobs} workers:")
    logging.info(tasks)
    with Pool(jobs) as p:
        p.starmap(collect_attacker_experience, tasks)

    end = time.time()
    elapsed = end - start
    logging.info(f"Time: {elapsed}s")




"""This file consist of the N-1 Tutor, that executes the actions similar to the N-1 Teacher.

"""
import logging
import time
from copy import deepcopy
from pathlib import Path
from typing import Optional, Union, Tuple, List
import numpy as np
import grid2op
from grid2op.Action import BaseAction
from grid2op.Observation import BaseObservation

from curriculumagent.common.utilities import simulate_action, find_best_line_to_reconnect, revert_topo
from curriculumagent.tutor.tutors.general_tutor import GeneralTutor, print_status


class NminusOneTutor(GeneralTutor):
    """The N-1 Tutor that searches through the actions, while checking whether the action is N-1 compatible.

    """

    def __init__(
            self,
            action_space: grid2op.Action.ActionSpace,
            action_space_file: Union[Path, List[Path]],
            do_nothing_threshold: Optional[float] = 0.9,
            best_action_threshold: Optional[float] = 0.99,
            rho_greedy_threshold: Optional[float] = 0.99,
            lines_to_check: Optional[Union[List[int], bool]] = [45, 56, 0, 9, 13, 14, 18, 23, 27, 39],
            return_status: Optional[bool] = True,
            revert_to_original_topo: Optional[bool] = False,
    ):
        """Simplified constructor method of the Tutor class

        The required actions are either a Path variable leading to a numpy array or a list with
        multiple entries (paths). If multiple entries are supplied, the tutor concatenates the actions to one
        combined list.

        Args:
            action_space: Action space object from Gird2Op environment.
            action_space_file: Either Numpy file with actions or List with multiple actions.
                If multiple action sets are supplied, the first is used for the N-1 search.
            do_nothing_threshold: Threshold, when the do nothing action is not sufficient.
            rho_greedy_threshold: Threshold, between do_nothing_threshold and best_action_threshold.
                If the max_rho is below this threshold, we check for N-1 search possibility.
            best_action_threshold: Threshold, when the collected action is sufficient and the search can
            be stopped. This is used for the greedy actions.
            lines_to_check: Indicator, whether the lines should be checked for N-1 actions.
            Can either be a boolean (to check all or none) or a specific list containing the
            different lines.
            return_status: Whether each step should be logged
            revert_to_original_topo: Boolean, whether to revert to original topology.

        Returns:
            None.

        """
        super().__init__(
            action_space=action_space,
            action_space_file=action_space_file,
            do_nothing_threshold=do_nothing_threshold,
            best_action_threshold=best_action_threshold,
            return_status=return_status,
        )

        if isinstance(lines_to_check, list):
            # List
            self.lines_to_attack = lines_to_check
            # Bool:
        elif lines_to_check:
            self.lines_to_attack = list(np.arange(action_space.n_line))
        else:
            self.lines_to_attack = []

        if (rho_greedy_threshold <= best_action_threshold) and (do_nothing_threshold <= rho_greedy_threshold):
            self.rho_greedy_threshold = rho_greedy_threshold
        else:
            logging.warning(
                "The provided rho_greedy_threshold is not inbetween the do_nothing_threshold"
                "and best_action_threshold. Set variable to default"
            )
            self.rho_greedy_threshold = 0.99

        self.revert_to_original_topo = revert_to_original_topo

    def act_with_id(self, observation: BaseObservation) -> Tuple[np.ndarray, int]:
        """In this methode, we execute the N-1 Action search.

        If no line is disconnected , we check for each action which would be the most stable N-1 result
        (i.e. which returns the min rho the most fatal line gets disconnected).
        Then this action is selected.
        If no action is found OR one line is already disconnected, we run the greedy action similar to
        the general_tutor.

        Note: If multiple action sets are provided, the first is used for the N-1 search.

        Args:
            observation: Grid2Op observation.

        Returns:
            Best action as numpy array as well as a tuple index for the action.

        """
        start_time = time.time()

        # Check for do nothing
        if observation.rho.max() < self.do_nothing_threshold:
            # secure, return "do nothing" in bus switches.

            if self.revert_to_original_topo:
                act = revert_topo(self.action_space, observation)
            else:
                act = np.zeros(self.action_space_size)

            return act, -1

        # Take lower values of either rho max or do nothing rho max
        obs_dn, _, _, _ = observation.simulate(self.action_space({}))

        min_rho = observation.rho.max()
        old_rho_max = min_rho.copy()
        action_chosen = None
        best_action_index = -1
        idx = None

        # N-1 Search IF no disconnected line ! or/and limit breached.
        if self.__n_minus_one_search_possible(obs=observation):
            rho_max_max = np.inf
            for idx, action_array in self.actions[0].items():
                act = self.action_space.from_vect(action_array)
                current_rho_max_max = self.calculate_attacked_max_rho(obs=observation, action=act)
                if current_rho_max_max < rho_max_max:
                    rho_max_max = current_rho_max_max
                    action_chosen = action_array
                    best_action_index = idx

            if action_chosen is not None:
                # Now look at the best option and again simulate the effect of the action prior to returning it.
                rho_max, valid_action = simulate_action(
                    action_vect=action_chosen, action_space=self.action_space, obs=observation
                )
                if rho_max < min_rho and valid_action:
                    if self.return_status:
                        print_status(observation, best_action_index, observation.rho.max(), rho_max, start_time)
                    logging.info(f"Gathered a N-1 Action with the id{idx}")
                    return action_chosen, best_action_index

        action_chosen = None

        for actions in self.actions:
            # Run through action set
            for idx, action_array in actions.items():
                obs_sim, valid_action = simulate_action(
                    action_vect=action_array, action_space=self.action_space, obs=observation
                )
                if not valid_action:
                    continue

                # Simulate action (even though it might be illegal for tuple or triple action)
                if obs_sim < min_rho:
                    min_rho = obs_sim
                    action_chosen = action_array
                    best_action_index = idx

            if min_rho <= self.action_threshold:
                break

        if self.return_status:
            print_status(observation, best_action_index, old_rho_max, min_rho, start_time)

        if action_chosen is not None:
            out = action_chosen, best_action_index

        else:
            out = np.zeros(self.action_space_size), -1

        return out

    def calculate_attacked_max_rho(self, obs: BaseObservation, action: BaseAction) -> float:
        """Method to run through all lines and check the maximum rho if the lines is disconnected.
        Note that if the game is done, the rho will be set to np.inf
        This is to measure how robust the supplied action is against the attacks, the lower, the better.

        Args:
            obs: Observation of the Grid2Op Env.
            action: Current selected action.

        Returns:
            The maximum rho of the action.

        """
        p_max_max = 0.0
        for line in self.lines_to_attack:
            new_line_status_array = np.zeros(obs.rho.shape, dtype=int)
            new_line_status_array[line] = -1
            disconnect_action = deepcopy(action)
            disconnect_action.update({"set_line_status": new_line_status_array})
            obs_new, _, done, _ = obs.simulate(disconnect_action)
            rho_max = obs_new.rho.max()
            if rho_max > p_max_max and done is False:
                p_max_max = rho_max
            if done:
                p_max_max = np.inf

        return p_max_max

    def __n_minus_one_search_possible(self, obs: grid2op.Observation.BaseObservation) -> bool:
        """Small private method, that checks whether the N-1 Search is actually possible.

        Args:
            obs: Current observation.

        Returns:
             Boolean, whether the N-1 Search can be computed.

        """
        search_possible = True

        #  Check if below threshold:
        if obs.rho.max() > self.rho_greedy_threshold:
            search_possible = False
        # Check if any lines are disconnected:
        if not all(obs.line_status):
            search_possible = False
        # Check if actually any lines can be checked:
        if len(self.lines_to_attack) == 0:
            search_possible = False
        return search_possible
